import logging
from blatann.event_type import EventSource, Event
from blatann.nrf import nrf_types, nrf_events
from blatann.waitables.event_waitable import EventWaitable
from blatann.exceptions import InvalidStateException
from blatann.event_args import EventArgs

logger = logging.getLogger(__name__)


class GattcWriteCompleteEventArgs(EventArgs):
    def __init__(self, handle, status, data):
        self.handle = handle
        self.status = status
        self.data = data


class GattcWriter(object):
    """
    Class which implements the state machine for writing a value to a peripheral's attribute
    """
    _WRITE_OVERHEAD = 3       # Number of bytes per MTU that are overhead for the write operation
    _LONG_WRITE_OVERHEAD = 5  # Number of bytes per MTU that are overhead for the long write operations

    def __init__(self, ble_device, peer):
        """
        :type ble_device: blatann.device.BleDevice
        :type peer: blatann.peer.Peer
        """
        self.ble_device = ble_device
        self.peer = peer
        self._on_write_complete = EventSource("On Write Complete", logger)
        self._busy = False
        self._data = ""
        self._handle = 0x0000
        self._offset = 0
        self.peer.driver_event_subscribe(self._on_write_response, nrf_events.GattcEvtWriteResponse)
        self._len_bytes_written = 0

    @property
    def on_write_complete(self):
        """
        Event that is emitted when a write completes on an attribute handler

        Handler args: (int attribute_handle, gatt.GattStatusCode, bytearray data_written)

        :return: an Event which can have handlers registered to and deregistered from
        :rtype: Event
        """
        return self._on_write_complete

    def write(self, handle, data):
        """
        Writes data to the attribute at the handle provided. Can only write to a single attribute at a time.
        If a write is in progress, raises an InvalidStateException

        :param handle: The attribute handle to write
        :param data: The data to write
        :return: A Waitable that will fire when the write finishes. see on_write_complete for the values returned from the waitable
        :rtype: EventWaitable
        """
        if self._busy:
            raise InvalidStateException("Gattc Writer is busy")
        if len(data) == 0:
            raise ValueError("Data must be at least one byte")
        self._offset = 0
        self._handle = handle
        self._data = data
        logger.debug("Starting write to handle {}, len: {}".format(self._handle, len(self._data)))
        self._write_next_chunk()
        self._busy = True
        return EventWaitable(self.on_write_complete)

    def _write_next_chunk(self):
        flags = nrf_types.BLEGattExecWriteFlag.unused
        if self._offset != 0 or len(self._data) > (self.peer.mtu_size - self._WRITE_OVERHEAD):
            write_operation = nrf_types.BLEGattWriteOperation.prepare_write_req
            self._len_bytes_written = self.peer.mtu_size - self._LONG_WRITE_OVERHEAD
            self._len_bytes_written = min(self._len_bytes_written, len(self._data)-self._offset)
            if self._len_bytes_written <= 0:
                write_operation = nrf_types.BLEGattWriteOperation.execute_write_req
                flags = nrf_types.BLEGattExecWriteFlag.prepared_write
        else:
            # Can write it all in a single
            write_operation = nrf_types.BLEGattWriteOperation.write_req
            self._len_bytes_written = len(self._data)

        data_to_write = self._data[self._offset:self._offset+self._len_bytes_written]
        write_params = nrf_types.BLEGattcWriteParams(write_operation, flags,
                                                     self._handle, data_to_write, self._offset)
        logger.debug("Writing chunk: handle: {}, offset: {}, len: {}, op: {}".format(self._handle, self._offset,
                                                                                     len(data_to_write), write_operation))
        self.ble_device.ble_driver.ble_gattc_write(self.peer.conn_handle, write_params)

    def _on_write_response(self, driver, event):
        """
        Handler for GattcEvtWriteResponse

        :type event: nrf_events.GattcEvtWriteResponse
        """
        if event.conn_handle != self.peer.conn_handle:
            return
        if event.attr_handle != self._handle and event.write_op != nrf_types.BLEGattWriteOperation.execute_write_req:
            return
        if event.status != nrf_events.BLEGattStatusCode.success:
            self._complete(event.status)
            return

        # Write successful, update offset and check operation
        self._offset += self._len_bytes_written

        if event.write_op in [nrf_types.BLEGattWriteOperation.write_req, nrf_types.BLEGattWriteOperation.execute_write_req]:
            # Completed successfully
            self._complete()
        elif event.write_op == nrf_types.BLEGattWriteOperation.prepare_write_req:
            # Write next chunk (or execute if complete)
            self._write_next_chunk()
        else:
            logger.error("Got unknown write operation: {}".format(event))
            self._complete(nrf_types.BLEGattStatusCode.unknown)

    def _complete(self, status=nrf_events.BLEGattStatusCode.success):
        self._busy = False
        self._on_write_complete.notify(self, GattcWriteCompleteEventArgs(self._handle, status, self._data))



from enum import IntEnum

from blatann.nrf.nrf_types import *
from blatann.nrf.nrf_dll_load import driver
import blatann.nrf.nrf_driver_types as util


class BLEEvent(object):
    evt_id = None

    def __init__(self, conn_handle):
        self.conn_handle = conn_handle

    def __str__(self):
        return self.__repr__()

    def _repr_format(self, *args, **kwargs):
        kwargs["conn_handle"] = self.conn_handle
        items = list(args) + ["{}={}".format(k, v) for k, v in kwargs.items()]
        inner = ", ".join(items)
        return "{}({})".format(self.__class__.__name__, inner)


class EvtUserMemoryRequest(BLEEvent):
    evt_id = driver.BLE_EVT_USER_MEM_REQUEST

    def __init__(self, conn_handle, request_type):
        super(EvtUserMemoryRequest, self).__init__(conn_handle)
        self.type = request_type

    @classmethod
    def from_c(cls, event):
        return cls(event.evt.common_evt.conn_handle, event.evt.common_evt.params.user_mem_request.type)

    def __repr__(self):
        return "{}(conn_handle={!r}, type={!r})".format(self.__class__.__name__, self.conn_handle, self.type)



from enum import Enum, IntEnum, IntFlag
from blatann.nrf.nrf_dll_load import driver

"""
Generic Enums
"""


class BLEHci(Enum):
    success = driver.BLE_HCI_STATUS_CODE_SUCCESS
    unknown_btle_command = driver.BLE_HCI_STATUS_CODE_UNKNOWN_BTLE_COMMAND
    unknown_connection_identifier = driver.BLE_HCI_STATUS_CODE_UNKNOWN_CONNECTION_IDENTIFIER
    authentication_failure = driver.BLE_HCI_AUTHENTICATION_FAILURE
    pin_or_key_missing = driver.BLE_HCI_STATUS_CODE_PIN_OR_KEY_MISSING
    memory_capacity_exceeded = driver.BLE_HCI_MEMORY_CAPACITY_EXCEEDED
    connection_timeout = driver.BLE_HCI_CONNECTION_TIMEOUT
    command_disallowed = driver.BLE_HCI_STATUS_CODE_COMMAND_DISALLOWED
    invalid_btle_command_parameters = driver.BLE_HCI_STATUS_CODE_INVALID_BTLE_COMMAND_PARAMETERS
    remote_user_terminated_connection = driver.BLE_HCI_REMOTE_USER_TERMINATED_CONNECTION
    remote_dev_termination_due_to_low_resources = driver.BLE_HCI_REMOTE_DEV_TERMINATION_DUE_TO_LOW_RESOURCES
    remote_dev_termination_due_to_power_off = driver.BLE_HCI_REMOTE_DEV_TERMINATION_DUE_TO_POWER_OFF
    local_host_terminated_connection = driver.BLE_HCI_LOCAL_HOST_TERMINATED_CONNECTION
    unsupported_remote_feature = driver.BLE_HCI_UNSUPPORTED_REMOTE_FEATURE
    invalid_lmp_parameters = driver.BLE_HCI_STATUS_CODE_INVALID_LMP_PARAMETERS
    unspecified_error = driver.BLE_HCI_STATUS_CODE_UNSPECIFIED_ERROR
    lmp_response_timeout = driver.BLE_HCI_STATUS_CODE_LMP_RESPONSE_TIMEOUT
    lmp_transaction_collision = driver.BLE_HCI_STATUS_CODE_LMP_ERROR_TRANSACTION_COLLISION
    lmp_pdu_not_allowed = driver.BLE_HCI_STATUS_CODE_LMP_PDU_NOT_ALLOWED
    instant_passed = driver.BLE_HCI_INSTANT_PASSED
    pairintg_with_unit_key_unsupported = driver.BLE_HCI_PAIRING_WITH_UNIT_KEY_UNSUPPORTED
    differen_transaction_collision = driver.BLE_HCI_DIFFERENT_TRANSACTION_COLLISION
    controller_busy = driver.BLE_HCI_CONTROLLER_BUSY
    conn_interval_unacceptable = driver.BLE_HCI_CONN_INTERVAL_UNACCEPTABLE
    parameter_out_of_mandatory_range = driver.BLE_HCI_PARAMETER_OUT_OF_MANDATORY_RANGE
    directed_advertiser_timeout = driver.BLE_HCI_DIRECTED_ADVERTISER_TIMEOUT
    conn_terminated_due_to_mic_failure = driver.BLE_HCI_CONN_TERMINATED_DUE_TO_MIC_FAILURE
    conn_failed_to_be_established = driver.BLE_HCI_CONN_FAILED_TO_BE_ESTABLISHED


class NrfError(Enum):
    # Generic errors
    success = driver.NRF_SUCCESS
    svc_handler_missing = driver.NRF_ERROR_SVC_HANDLER_MISSING
    softdevice_not_enabled = driver.NRF_ERROR_SOFTDEVICE_NOT_ENABLED
    internal = driver.NRF_ERROR_INTERNAL
    no_mem = driver.NRF_ERROR_NO_MEM
    not_found = driver.NRF_ERROR_NOT_FOUND
    not_supported = driver.NRF_ERROR_NOT_SUPPORTED
    invalid_param = driver.NRF_ERROR_INVALID_PARAM
    invalid_state = driver.NRF_ERROR_INVALID_STATE
    invalid_length = driver.NRF_ERROR_INVALID_LENGTH
    invalid_flags = driver.NRF_ERROR_INVALID_FLAGS
    invalid_data = driver.NRF_ERROR_INVALID_DATA
    data_size = driver.NRF_ERROR_DATA_SIZE
    timeout = driver.NRF_ERROR_TIMEOUT
    null = driver.NRF_ERROR_NULL
    forbidden = driver.NRF_ERROR_FORBIDDEN
    invalid_addr = driver.NRF_ERROR_INVALID_ADDR
    busy = driver.NRF_ERROR_BUSY
    conn_count = driver.NRF_ERROR_CONN_COUNT
    resources = driver.NRF_ERROR_RESOURCES

    # sdm_lfclk_source_unknown = driver.NRF_ERROR_SDM_LFCLK_SOURCE_UNKNOWN
    # sdm_incorrect_interrupt_configuration = driver.NRF_ERROR_SDM_INCORRECT_INTERUUPT_CONFIGURATION
    # sdm_incorrect_clenr0 = driver.NRF_ERROR_SDM_INCORRECT_CLENR0
    # soc_mutex_already_taken = driver.NRF_ERROR_SOC_MUTEX_ALREADY_TAKEN
    # soc_nvic_interrupt_not_available = driver.NRF_ERROR_SOC_NVIC_INTERRUPT_NOT_AVAILABLE
    # soc_nvic_interrupt_priority_not_allowed = driver.NRF_ERROR_SOC_NVIC_INTERRUPT_PRIORITY_NOT_ALLOWED
    # soc_nvic_should_not_return = driver.NRF_ERROR_SOC_NVIC_SHOULD_NOT_RETURN
    # soc_power_mode_unknown = driver.NRF_ERROR_SOC_POWER_MODE_UNKNOWN
    # soc_power_pof_threshold_unknown = driver.NRF_ERROR_SOC_POWER_POF_THRESHOLD_UNKNOWN
    # soc_power_off_should_not_return = driver.NRF_ERROR_SOC_POWER_OFF_SHOULD_NOT_RETURN
    # soc_rand_not_enough_values = driver.NRF_ERROR_SOC_RAND_NOT_ENOUGH_VALUES
    # soc_ppi_invalid_channel = driver.NRF_ERROR_SOC_PPI_INVALID_CHANNEL
    # soc_ppi_invalid_group = driver.NRF_ERROR_SOC_PPI_INVALID_GROUP
    # ble_not_enabled = driver.BLE_ERROR_NOT_ENABLED
    # ble_invalid_conn_handle = driver.BLE_ERROR_INVALID_CONN_HANDLE
    # ble_invalid_attr_handle = driver.BLE_ERROR_INVALID_ATTR_HANDLE
    # ble_invalid_role = driver.BLE_ERROR_INVALID_ROLE
    # ble_gap_uuid_list_mismatch = driver.BLE_ERROR_GAP_UUID_LIST_MISMATCH
    # ble_gap_discoverable_with_whitelist = driver.BLE_ERROR_GAP_DISCOVERABLE_WITH_WHITELIST
    # ble_gap_invalid_ble_addr = driver.BLE_ERROR_GAP_INVALID_BLE_ADDR
    # ble_gap_whitelist_in_use = driver.BLE_ERROR_GAP_WHITELIST_IN_USE
    # ble_gap_device_identities_in_use = driver.BLE_ERROR_GAP_DEVICE_IDENTITIES_IN_USE
    # ble_gap_device_identities_duplicate = driver.BLE_ERROR_GAP_DEVICE_IDENTITIES_DUPLICATE
    # ble_gatts_invalid_attr_type = driver.BLE_ERROR_GATTS_INVALID_ATTR_TYPE
    # ble_gatts_sys_attr_missing = driver.BLE_ERROR_GATTS_SYS_ATTR_MISSING
    # ble_gattc_proc_not_permitted = driver.BLE_ERROR_GATTC_PROC_NOT_PERMITTED


"""
Gap Enums
"""

BLE_CONN_HANDLE_INVALID = driver.BLE_CONN_HANDLE_INVALID


class BLEGapAdvType(IntEnum):
    connectable_undirected = driver.BLE_GAP_ADV_TYPE_ADV_IND
    connectable_directed = driver.BLE_GAP_ADV_TYPE_ADV_DIRECT_IND
    scanable_undirected = driver.BLE_GAP_ADV_TYPE_ADV_SCAN_IND
    non_connectable_undirected = driver.BLE_GAP_ADV_TYPE_ADV_NONCONN_IND
    scan_response = 4


class BLEGapRoles(IntEnum):
    invalid = driver.BLE_GAP_ROLE_INVALID
    periph = driver.BLE_GAP_ROLE_PERIPH
    central = driver.BLE_GAP_ROLE_CENTRAL


class BLEGapTimeoutSrc(IntEnum):
    advertising = driver.BLE_GAP_TIMEOUT_SRC_ADVERTISING
    scan = driver.BLE_GAP_TIMEOUT_SRC_SCAN
    conn = driver.BLE_GAP_TIMEOUT_SRC_CONN


class BLEGapPhy(IntFlag):
    auto = driver.BLE_GAP_PHY_AUTO
    one_mbps = driver.BLE_GAP_PHY_1MBPS
    two_mbps = driver.BLE_GAP_PHY_2MBPS
    coded = driver.BLE_GAP_PHY_CODED


"""
SMP Enums
"""


class BLEGapIoCaps(IntEnum):
    DISPLAY_ONLY = driver.BLE_GAP_IO_CAPS_DISPLAY_ONLY
    DISPLAY_YESNO = driver.BLE_GAP_IO_CAPS_DISPLAY_YESNO
    KEYBOARD_ONLY = driver.BLE_GAP_IO_CAPS_KEYBOARD_ONLY
    NONE = driver.BLE_GAP_IO_CAPS_NONE
    KEYBOARD_DISPLAY = driver.BLE_GAP_IO_CAPS_KEYBOARD_DISPLAY


class BLEGapAuthKeyType(IntEnum):
    NONE = driver.BLE_GAP_AUTH_KEY_TYPE_NONE
    OOB = driver.BLE_GAP_AUTH_KEY_TYPE_OOB
    PASSKEY = driver.BLE_GAP_AUTH_KEY_TYPE_PASSKEY


class BLEGapSecStatus(IntEnum):
    success = driver.BLE_GAP_SEC_STATUS_SUCCESS
    timeout = driver.BLE_GAP_SEC_STATUS_TIMEOUT
    pdu_invalid = driver.BLE_GAP_SEC_STATUS_PDU_INVALID
    passkey_entry_failed = driver.BLE_GAP_SEC_STATUS_PASSKEY_ENTRY_FAILED
    oob_not_available = driver.BLE_GAP_SEC_STATUS_OOB_NOT_AVAILABLE
    auth_req = driver.BLE_GAP_SEC_STATUS_AUTH_REQ
    confirm_value = driver.BLE_GAP_SEC_STATUS_CONFIRM_VALUE
    pairing_not_supp = driver.BLE_GAP_SEC_STATUS_PAIRING_NOT_SUPP
    enc_key_size = driver.BLE_GAP_SEC_STATUS_ENC_KEY_SIZE
    smp_cmd_unsupported = driver.BLE_GAP_SEC_STATUS_SMP_CMD_UNSUPPORTED
    unspecified = driver.BLE_GAP_SEC_STATUS_UNSPECIFIED
    repeated_attempts = driver.BLE_GAP_SEC_STATUS_REPEATED_ATTEMPTS
    invalid_params = driver.BLE_GAP_SEC_STATUS_INVALID_PARAMS
    dhkey_failure = driver.BLE_GAP_SEC_STATUS_DHKEY_FAILURE
    num_comp_failure = driver.BLE_GAP_SEC_STATUS_NUM_COMP_FAILURE
    br_edr_in_prog = driver.BLE_GAP_SEC_STATUS_BR_EDR_IN_PROG
    x_trans_key_disallowed = driver.BLE_GAP_SEC_STATUS_X_TRANS_KEY_DISALLOWED


"""
GATT Enums
"""


class BLEGattWriteOperation(Enum):
    invalid = driver.BLE_GATT_OP_INVALID
    write_req = driver.BLE_GATT_OP_WRITE_REQ
    write_cmd = driver.BLE_GATT_OP_WRITE_CMD
    signed_write_cmd = driver.BLE_GATT_OP_SIGN_WRITE_CMD
    prepare_write_req = driver.BLE_GATT_OP_PREP_WRITE_REQ
    execute_write_req = driver.BLE_GATT_OP_EXEC_WRITE_REQ


class BLEGattHVXType(Enum):
    invalid = driver.BLE_GATT_HVX_INVALID
    notification = driver.BLE_GATT_HVX_NOTIFICATION
    indication = driver.BLE_GATT_HVX_INDICATION


class BLEGattStatusCode(Enum):
    success = driver.BLE_GATT_STATUS_SUCCESS
    unknown = driver.BLE_GATT_STATUS_UNKNOWN
    invalid = driver.BLE_GATT_STATUS_ATTERR_INVALID
    invalid_handle = driver.BLE_GATT_STATUS_ATTERR_INVALID_HANDLE
    read_not_permitted = driver.BLE_GATT_STATUS_ATTERR_READ_NOT_PERMITTED
    write_not_permitted = driver.BLE_GATT_STATUS_ATTERR_WRITE_NOT_PERMITTED
    invalid_pdu = driver.BLE_GATT_STATUS_ATTERR_INVALID_PDU
    insuf_authentication = driver.BLE_GATT_STATUS_ATTERR_INSUF_AUTHENTICATION
    request_not_supported = driver.BLE_GATT_STATUS_ATTERR_REQUEST_NOT_SUPPORTED
    invalid_offset = driver.BLE_GATT_STATUS_ATTERR_INVALID_OFFSET
    insuf_authorization = driver.BLE_GATT_STATUS_ATTERR_INSUF_AUTHORIZATION
    prepare_queue_full = driver.BLE_GATT_STATUS_ATTERR_PREPARE_QUEUE_FULL
    attribute_not_found = driver.BLE_GATT_STATUS_ATTERR_ATTRIBUTE_NOT_FOUND
    attribute_not_long = driver.BLE_GATT_STATUS_ATTERR_ATTRIBUTE_NOT_LONG
    insuf_enc_key_size = driver.BLE_GATT_STATUS_ATTERR_INSUF_ENC_KEY_SIZE
    invalid_att_val_length = driver.BLE_GATT_STATUS_ATTERR_INVALID_ATT_VAL_LENGTH
    unlikely_error = driver.BLE_GATT_STATUS_ATTERR_UNLIKELY_ERROR
    insuf_encryption = driver.BLE_GATT_STATUS_ATTERR_INSUF_ENCRYPTION
    unsupported_group_type = driver.BLE_GATT_STATUS_ATTERR_UNSUPPORTED_GROUP_TYPE
    insuf_resources = driver.BLE_GATT_STATUS_ATTERR_INSUF_RESOURCES
    rfu_range1_begin = driver.BLE_GATT_STATUS_ATTERR_RFU_RANGE1_BEGIN
    rfu_range1_end = driver.BLE_GATT_STATUS_ATTERR_RFU_RANGE1_END
    app_begin = driver.BLE_GATT_STATUS_ATTERR_APP_BEGIN
    app_end = driver.BLE_GATT_STATUS_ATTERR_APP_END
    rfu_range2_begin = driver.BLE_GATT_STATUS_ATTERR_RFU_RANGE2_BEGIN
    rfu_range2_end = driver.BLE_GATT_STATUS_ATTERR_RFU_RANGE2_END
    rfu_range3_begin = driver.BLE_GATT_STATUS_ATTERR_RFU_RANGE3_BEGIN
    rfu_range3_end = driver.BLE_GATT_STATUS_ATTERR_RFU_RANGE3_END
    cps_cccd_config_error = driver.BLE_GATT_STATUS_ATTERR_CPS_CCCD_CONFIG_ERROR
    cps_proc_alr_in_prog = driver.BLE_GATT_STATUS_ATTERR_CPS_PROC_ALR_IN_PROG
    cps_out_of_range = driver.BLE_GATT_STATUS_ATTERR_CPS_OUT_OF_RANGE


class BLEGattExecWriteFlag(Enum):
    prepared_cancel = driver.BLE_GATT_EXEC_WRITE_FLAG_PREPARED_CANCEL
    prepared_write = driver.BLE_GATT_EXEC_WRITE_FLAG_PREPARED_WRITE
    unused = 0x00

"""
GATTC Enums
"""


"""
GATTS Enums
"""


class BLEGattsWriteOperation(Enum):
    invalid = driver.BLE_GATTS_OP_INVALID
    write_req = driver.BLE_GATTS_OP_WRITE_REQ
    write_cmd = driver.BLE_GATTS_OP_WRITE_CMD
    sign_write_cmd = driver.BLE_GATTS_OP_SIGN_WRITE_CMD
    prep_write_req = driver.BLE_GATTS_OP_PREP_WRITE_REQ
    exec_write_req_cancel = driver.BLE_GATTS_OP_EXEC_WRITE_REQ_CANCEL
    exec_write_req_now = driver.BLE_GATTS_OP_EXEC_WRITE_REQ_NOW



from enum import IntEnum
from blatann.services import ble_data_types


# See https://www.bluetooth.com/specifications/gatt/viewer?attributeXmlFile=org.bluetooth.service.glucose.xml
# For more info about the data types and values defined here


class GlucoseConcentrationUnits(IntEnum):
    """
    The concentration units available for reporting glucose levels
    """
    kg_per_liter = 0
    mol_per_liter = 1


class GlucoseType(IntEnum):
    """
    The glucose types available
    """
    capillary_whole_blood = 1
    capillary_plasma = 2
    venous_whole_blood = 3
    venous_plasma = 4
    arterial_whole_blood = 5
    arterial_plasma = 6
    undetermined_whole_blood = 7
    undetermined_plasma = 8
    interstitial_fluid = 9
    control_solution = 10


class SampleLocation(IntEnum):
    """
    Location which the blood sample was taken
    """
    finger = 1
    alternate_test_site = 2
    earlobe = 3
    control_solution = 4
    unknown = 15


class MedicationUnits(IntEnum):
    """
    Available units to report medication values in
    """
    milligrams = 0
    milliliters = 1


class CarbohydrateType(IntEnum):
    """
    The type of carbohydrate consumed by the user
    """
    breakfast = 1
    lunch = 2
    dinner = 3
    snack = 4
    drink = 5
    supper = 6
    brunch = 7


class MealType(IntEnum):
    """
    The type of meal consumed
    """
    preprandial = 1   # Before Meal
    postprandial = 2  # After Meal
    fasting = 3
    casual = 4        # Snacks, drinks, etc.
    bedtime = 5


class TesterType(IntEnum):
    """
    Information about who tested the glucose levels
    """
    self = 1
    health_care_professional = 2
    lab_test = 3
    not_available = 15


class HealthStatus(IntEnum):
    """
    Current health status of the user
    """
    minor_issues = 1
    major_issues = 2
    during_menses = 3
    under_stress = 4
    normal = 5
    not_available = 15


class MedicationType(IntEnum):
    """
    Medication type consumed
    """
    rapid_acting_insulin = 1
    short_acting_insulin = 2
    intermediate_acting_insulin = 3
    long_acting_insulin = 4
    premixed_insulin = 5


class SensorStatusType(IntEnum):
    """
    The types of sensor statuses that can be communicated
    """
    battery_low = 0
    sensor_malfunction = 1
    sample_size_insufficient = 2
    strip_insertion_error = 3
    incorrect_strip_type = 4
    result_above_range = 5
    result_below_range = 6
    sensor_temp_high = 7
    sensor_temp_low = 8
    sensor_read_interrupted = 9
    general_device_fault = 10
    time_fault = 11


class SensorStatus(ble_data_types.Bitfield):
    """
    Class which holds the current sensor status information
    """
    bitfield_width = 16
    bitfield_enum = SensorStatusType

    def __init__(self, *sensor_statuses):
        """
        :param sensor_statuses: The list of SensorStatusTypes that are currently active on the device
        :type sensor_statuses: SensorStatusType
        """
        # Field names must match enum names exactly
        self.battery_low = SensorStatusType.battery_low in sensor_statuses
        self.sensor_malfunction = SensorStatusType.sensor_malfunction in sensor_statuses
        self.sample_size_insufficient = SensorStatusType.sample_size_insufficient in sensor_statuses
        self.strip_insertion_error = SensorStatusType.strip_insertion_error in sensor_statuses
        self.incorrect_strip_type = SensorStatusType.incorrect_strip_type in sensor_statuses
        self.result_above_range = SensorStatusType.result_above_range in sensor_statuses
        self.result_below_range = SensorStatusType.result_below_range in sensor_statuses
        self.sensor_temp_high = SensorStatusType.sensor_temp_high in sensor_statuses
        self.sensor_temp_low = SensorStatusType.sensor_temp_low in sensor_statuses
        self.sensor_read_interrupted = SensorStatusType.sensor_read_interrupted in sensor_statuses
        self.general_device_fault = SensorStatusType.general_device_fault in sensor_statuses
        self.time_fault = SensorStatusType.time_fault in sensor_statuses

        super(SensorStatus, self).__init__()


class GlucoseFeatureType(IntEnum):
    """
    Enumeration of the supported feature types to be reported
    using the Feature characteristic
    """
    low_battery_detection = 0
    sensor_malfunction_detection = 1
    sensor_sample_size = 2
    strip_insertion_error_detection = 3
    strip_type_error_detection = 4
    sensor_result_high_low_detection = 5
    sensor_temp_high_low_detection = 6
    sensor_read_interrupt_detection = 7
    general_device_fault = 8
    time_fault = 9
    multiple_bond = 10


class GlucoseFeatures(ble_data_types.Bitfield):
    """
    Class which holds the features of the glucose sensor and is reported to over bluetooth.
    This is the class used for the Feature characteristic.
    """
    bitfield_width = 16
    bitfield_enum = GlucoseFeatureType

    def __init__(self, *supported_features):
        """
        :param supported_features: The features that are supported by the sensor
        :type supported_features: GlucoseFeatureType
        """
        # Field names must match enum names exactly
        self.low_battery_detection = GlucoseFeatureType.low_battery_detection in supported_features
        self.sensor_malfunction_detection = GlucoseFeatureType.sensor_malfunction_detection in supported_features
        self.sensor_sample_size = GlucoseFeatureType.sensor_sample_size in supported_features
        self.strip_insertion_error_detection = GlucoseFeatureType.strip_insertion_error_detection in supported_features
        self.strip_type_error_detection = GlucoseFeatureType.strip_type_error_detection in supported_features
        self.sensor_result_high_low_detection = GlucoseFeatureType.sensor_result_high_low_detection in supported_features
        self.sensor_temp_high_low_detection = GlucoseFeatureType.sensor_temp_high_low_detection in supported_features
        self.sensor_read_interrupt_detection = GlucoseFeatureType.sensor_read_interrupt_detection in supported_features
        self.general_device_fault = GlucoseFeatureType.general_device_fault in supported_features
        self.time_fault = GlucoseFeatureType.time_fault in supported_features
        self.multiple_bond = GlucoseFeatureType.multiple_bond in supported_features

        super(GlucoseFeatures, self).__init__()


class _MeasurementFlags(ble_data_types.Bitfield):
    """
    Bitfield used in the GlucoseMeasurement struct which defines
    which fields are present in the message
    """
    class Bits(IntEnum):
        time_offset_present = 0
        sample_present = 1
        concentration_units = 2
        sensor_status = 3
        has_context = 4

    bitfield_width = 8
    bitfield_enum = Bits

    def __init__(self):
        self.time_offset_present = False
        self.sample_present = False
        self.concentration_units = GlucoseConcentrationUnits.kg_per_liter
        self.sensor_status = False
        self.has_context = False

        super(_MeasurementFlags, self).__init__()


class GlucoseSample(ble_data_types.BleCompoundDataType):
    """
    Holds the info about a glucose sample to be reported through the Glucose Measurement characteristic
    """
    def __init__(self, glucose_type, sample_location, value, units=GlucoseConcentrationUnits.kg_per_liter):
        """
        :param glucose_type: The type of blood the glucose sample is from
        :type glucose_type: GlucoseType
        :param sample_location: the body location the sample was taken from
        :type sample_location: SampleLocation
        :param value: The glucose reading taken, in units specified by the units parameter
        :type value: float
        :param units: The units of the glucose sample
        :type units: GlucoseConcentrationUnits
        """
        self.type = glucose_type
        self.value = value
        self.sample_location = sample_location
        # Units are specified in a separate bitfield, not encoded or decoded
        self.units = units

    def encode(self):
        stream = ble_data_types.BleDataStream()
        stream.encode(ble_data_types.SFloat, self.value)
        stream.encode(ble_data_types.DoubleNibble, [self.type, self.sample_location])
        return stream

    @classmethod
    def decode(cls, stream):
        value = stream.decode(ble_data_types.SFloat)
        glucose_type, location = stream.decode(ble_data_types.DoubleNibble)

        return GlucoseSample(glucose_type, location, value)

    def __repr__(self):
        return "{}({}, {}, {}, {})".format(self.__class__.__name__, str(self.type), self.value,
                                           str(self.units), str(self.sample_location))


class GlucoseMeasurement(ble_data_types.BleCompoundDataType):
    """
    Represents a single measurement taken and can be reported over BLE
    """
    def __init__(self, sequence_number, measurement_time, time_offset_minutes=None,
                 sample=None, sensor_status=None, context=None):
        """
        :param sequence_number: the sequence number of the measurement. Mandatory.
        :type sequence_number: int
        :param measurement_time: The time at which the measurement occurred. Must be a datetime struct. Mandatory.
        :type measurement_time: datetime.datetime
        :param time_offset_minutes: The time offset of the measurement, in minutes. Optional.
        :type time_offset_minutes: int
        :param sample: The blood glucose reading. Optional.
        :type sample: GlucoseSample
        :param sensor_status: The status of the glucose sensor. Optional.
        :type sensor_status: SensorStatus
        :param context: The glucose context to be reported with the measurement. Optional
        :type context: GlucoseContext
        """
        self.sequence_number = sequence_number
        self.measurement_time = measurement_time
        self.time_offset_minutes = time_offset_minutes
        self.sample = sample
        self.sensor_status = sensor_status
        self.context = context

    def encode(self):
        stream = ble_data_types.BleDataStream()

        flags = _MeasurementFlags()
        flags.time_offset_present = self.time_offset_minutes is not None
        flags.sample_present = self.sample is not None
        if flags.sample_present:
            flags.concentration_units = int(self.sample.units)
        flags.sensor_status = self.sensor_status is not None
        flags.has_context = self.context is not None

        stream.encode(flags)
        stream.encode(ble_data_types.Uint16, self.sequence_number)
        stream.encode(ble_data_types.DateTime(self.measurement_time))
        stream.encode_if(flags.time_offset_present, ble_data_types.Int16, self.time_offset_minutes)
        stream.encode_if(flags.sample_present, self.sample)
        stream.encode_if(flags.sensor_status, self.sensor_status)

        return stream

    @classmethod
    def decode(cls, stream):
        flags = stream.decode(_MeasurementFlags)
        sequence_number = stream.decode(ble_data_types.Uint16)
        time = stream.decode(ble_data_types.DateTime)
        units = GlucoseConcentrationUnits(int(flags.concentration_units))
        has_context = flags.has_context

        time_offset = stream.decode_if(flags.time_offset_present, ble_data_types.Int16)
        reading = stream.decode_if(flags.sample_present, GlucoseSample)
        if reading:
            reading.units = units
        sensor_status = stream.decode_if(flags.sensor_status, SensorStatus)

        return GlucoseMeasurement(sequence_number, time, time_offset, reading, sensor_status, has_context)

    def __repr__(self):
        params = ["seq: {}".format(self.sequence_number), "time: {}".format(self.measurement_time)]
        if self.time_offset_minutes is not None:
            params.append("time offset: {}".format(self.time_offset_minutes))
        if self.sample:
            params.append(str(self.sample))
        if self.sensor_status:
            params.append(str(self.sensor_status))
        if self.context:
            params.append(str(self.context))

        return "{}({})".format(self.__class__.__name__, ", ".join(params))


class CarbsInfo(ble_data_types.BleCompoundDataType):
    """
    Holds information about the carbs consumed
    """
    def __init__(self, carbs_grams, carb_type):
        """
        :param carbs_grams: The amount of carbs consumed, in grams
        :type carbs_grams: float
        :param carb_type: The type of carbs consumed
        :type carb_type: CarbohydrateType
        """
        self.carbs_grams = carbs_grams
        self.carb_type = carb_type

    def encode(self):
        stream = ble_data_types.BleDataStream()
        stream.encode(ble_data_types.Uint8, self.carb_type)
        stream.encode(ble_data_types.SFloat, self.carbs_grams)
        return stream

    @classmethod
    def decode(cls, stream):
        carb_type, carbs_grams = stream.decode_multiple(ble_data_types.Uint8, ble_data_types.SFloat)
        return CarbsInfo(carb_type, carbs_grams)

    def __repr__(self):
        return "{}({}g, {})".format(self.__class__.__name__, self.carbs_grams, str(self.carb_type))


class ExerciseInfo(ble_data_types.BleCompoundDataType):
    """
    Holds information about the exercise performed with the glucose sample
    """
    # Special value which represents that the exercise duration was longer than what can be reported (uint16 max)
    EXERCISE_DURATION_OVERRUN = 65535

    def __init__(self, duration_seconds, intensity_percent):
        """
        :param duration_seconds: The duration of exercise, in seconds. Can only report up to 65534 seconds over BLE
        :type duration_seconds: int
        :param intensity_percent: The exercise intensity, expressed as a percentage
        :type intensity_percent: int
        """
        self.duration_seconds = duration_seconds
        self.intensity_percent = intensity_percent

    def encode(self):
        stream = ble_data_types.BleDataStream()

        # Clamp duration to max 16-bit, max value means overrun
        duration = max(self.duration_seconds, self.EXERCISE_DURATION_OVERRUN)
        stream.encode(ble_data_types.Uint16, duration)
        stream.encode(ble_data_types.Uint8, self.intensity_percent)
        return stream

    @classmethod
    def decode(cls, stream):
        duration, intensity = stream.decode_multiple(ble_data_types.Uint16, ble_data_types.Uint8)
        return ExerciseInfo(duration, intensity)

    def __repr__(self):
        return "{}({} seconds, {}% intensity)".format(self.__class__.__name__,
                                                      self.duration_seconds, self.intensity_percent)


class MedicationInfo(ble_data_types.BleCompoundDataType):
    """
    Holds information about the medication administered
    """
    def __init__(self, med_type, med_value, med_units=MedicationUnits.milligrams):
        """
        :param med_type: The type of medication administered
        :type med_type: MedicationType
        :param med_value: The amount of medication administered, expressed in units specified
        :type med_value: float
        :param med_units: The units of medication
        :type med_units: MedicationUnits
        """
        self.type = med_type
        self.value = med_value
        # Units are specified in a separate bitfield, not encoded or decoded
        self.units = med_units

    def encode(self):
        stream = ble_data_types.BleDataStream()
        stream.encode(ble_data_types.Uint8, self.type)
        stream.encode(ble_data_types.SFloat, self.value)
        return stream

    @classmethod
    def decode(cls, stream):
        med_type, med_value = stream.decode_multiple(ble_data_types.Uint8, ble_data_types.SFloat)
        return MedicationInfo(med_type, med_value)

    def __repr__(self):
        return "{}({}, {} {})".format(self.__class__.__name__, self.value, str(self.units), str(self.type))


class _GlucoseContextFlags(ble_data_types.Bitfield):
    """
    Bitfield used in the GlucoseContext struct which defines
    which fields are present in the message
    """
    class Bits(IntEnum):
        carb_present = 0
        meal_present = 1
        tester_health_present = 2
        exercise_present = 3
        medication_present = 4
        medication_units = 5
        hba1c_present = 6
        extended_flags_present = 7

    bitfield_width = 8
    bitfield_enum = Bits

    def __init__(self):
        self.carb_present = False
        self.meal_present = False
        self.tester_health_present = False
        self.exercise_present = False
        self.medication_present = False
        self.medication_units = MedicationUnits.milligrams
        self.hba1c_present = False
        self.extended_flags_present = False

        super(_GlucoseContextFlags, self).__init__()


class GlucoseContext(ble_data_types.BleCompoundDataType):
    """
    Class which holds the extra glucose context data associated with the glucose measurement
    """

    def __init__(self, sequence_number, carbs=None, meal_type=None, tester=None, health_status=None,
                 exercise=None, medication=None, hba1c_percent=None, extra_flags=None):
        """
        :param sequence_number: The sequence number that the context corresponds to. Must match the GlucoseMeasurement sequence number
        :type sequence_number: int
        :param carbs: Information about the carbs consumed. Optional.
        :type carbs: CarbsInfo
        :param meal_type: The type of meal the reading was taken with. Optional
        :type meal_type: MealType
        :param tester: Who tested the glucose levels. Optional, must be present if health_status is specified.
        :type tester: TesterType
        :param health_status: The health status of the patient. Optional, must be present if tester is specified.
        :type health_status: HealthStatus
        :param exercise: Information about the exercise performed at time of sample. Optional.
        :type exercise: ExerciseInfo
        :param medication: Information about the medication administered to the patient. Optional.
        :type medication: MedicationInfo
        :param hba1c_percent:
        :param extra_flags:
        """
        self.sequence_number = sequence_number
        self.carbs = carbs
        self.meal_type = meal_type
        self.tester = tester
        self.health_status = health_status
        self.exercise = exercise
        self.medication = medication
        self.hba1c_percent = hba1c_percent
        self.extra_flags = extra_flags

    def encode(self):
        stream = ble_data_types.BleDataStream()

        flags = _GlucoseContextFlags()
        flags.carb_present = self.carbs is not None
        flags.meal_present = self.meal_type is not None
        flags.tester_health_present = self.tester is not None
        flags.exercise_present = self.exercise is not None
        flags.medication_present = self.medication is not None
        if flags.medication_present:
            flags.medication_units = self.medication.units
        flags.hba1c_present = self.hba1c_percent is not None
        flags.extended_flags_present = self.extra_flags is not None

        stream.encode(flags)
        stream.encode(ble_data_types.Uint16, self.sequence_number)

        stream.encode_if(flags.extended_flags_present, ble_data_types.Uint8, self.extra_flags)
        stream.encode_if(flags.carb_present, self.carbs)

        stream.encode_if(flags.meal_present, ble_data_types.Uint8, self.meal_type)
        stream.encode_if(flags.tester_health_present, ble_data_types.DoubleNibble, [self.tester, self.health_status])

        stream.encode_if(flags.exercise_present, self.exercise)

        stream.encode_if(flags.medication_present, self.medication)
        stream.encode_if(flags.hba1c_present, ble_data_types.SFloat, self.hba1c_percent)

        return stream

    @classmethod
    def decode(cls, stream):
        """
        :type stream: ble_data_types.BleDataStream
        """
        flags = stream.decode(_GlucoseContextFlags)
        med_units = MedicationUnits(int(flags.medication_units))

        sequence_number = stream.decode(ble_data_types.Uint16)
        extended_flags = stream.decode_if(flags.extended_flags_present, ble_data_types.Uint8)
        carbs = stream.decode_if(flags.carb_present, CarbsInfo)
        meal_type = stream.decode_if(flags.meal_present, ble_data_types.Uint8)
        tester, health = stream.decode_if(flags.tester_health_present, ble_data_types.DoubleNibble)
        exercise = stream.decode_if(flags.exercise_present, ExerciseInfo)
        medication = stream.decode_if(flags.medication_present, MedicationInfo)
        if medication:
            medication.units = med_units
        hba1c = stream.decode_if(flags.hba1c_present, ble_data_types.SFloat)

        return GlucoseContext(sequence_number, carbs, meal_type, tester, health, exercise, medication, hba1c, extended_flags)

    def __repr__(self):
        params = ["seq: {}".format(self.sequence_number)]
        if self.carbs:
            params.append(str(self.carbs))
        if self.meal_type:
            params.append(str(self.meal_type))
        if self.tester:
            params.append(str(self.tester))
        if self.health_status:
            params.append(str(self.health_status))
        if self.exercise:
            params.append(str(self.exercise))
        if self.medication:
            params.append(str(self.medication))
        if self.hba1c_percent:
            params.append("hba1c: {}%".format(self.hba1c_percent))
        if self.extra_flags:
            params.append("extra_flags: {}".format(self.extra_flags))
        return "{}({})".format(self.__class__.__name__, ", ".join(params))



import logging
from blatann.nrf import nrf_events, nrf_types
from blatann import exceptions
from blatann.waitables.connection_waitable import ClientConnectionWaitable
from blatann.event_type import Event, EventSource
from blatann.gap.advertise_data import AdvertisingData, AdvertisingFlags


logger = logging.getLogger(__name__)


AdvertisingMode = nrf_types.BLEGapAdvType


class Advertiser(object):
    # Constant used to indicate that the BLE device should advertise indefinitely, until
    # connected to or stopped manually
    ADVERTISE_FOREVER = 0

    def __init__(self, ble_device, client, conn_tag=0):
        """
        :type ble_device: blatann.device.BleDevice
        :type client: blatann.peer.Client
        """
        self.ble_device = ble_device
        self._is_advertising = False
        self._auto_restart = False
        self.client = client
        self.ble_device.ble_driver.event_subscribe(self._handle_adv_timeout, nrf_events.GapEvtTimeout)
        self.client.on_disconnect.register(self._handle_disconnect)
        self.client.on_connect.register(self._handle_connect)
        self._on_advertising_timeout = EventSource("Advertising Timeout", logger)
        self._advertise_interval = 100
        self._timeout = self.ADVERTISE_FOREVER
        self._advertise_mode = AdvertisingMode.connectable_undirected
        self._conn_tag = conn_tag

    @property
    def on_advertising_timeout(self):
        """
        Event generated whenever advertising times out and finishes with no connections made
        Event args: None

        :return: an Event which can have handlers registered to and deregistered from
        :rtype: Event
        """
        return self._on_advertising_timeout

    @property
    def is_advertising(self):
        return self._is_advertising

    def set_advertise_data(self, advertise_data=AdvertisingData(), scan_response=AdvertisingData()):
        """
        Sets the advertising and scan response data which will be broadcasted to peers during advertising

        Note: BLE Restricts advertise and scan response data to an encoded length of 31 bytes each.
        Use AdvertisingData.check_encoded_length() to determine if the

        :param advertise_data: The advertise data to use
        :type advertise_data: AdvertisingData
        :param scan_response: The scan response data to use
        :type scan_response: AdvertisingData
        """
        adv_len, adv_pass = advertise_data.check_encoded_length()
        scan_len, scan_pass = advertise_data.check_encoded_length()

        if not adv_pass:
            raise exceptions.InvalidOperationException("Encoded Advertising data length is too long ({} bytes). "
                                                       "Max: {} bytes".format(adv_len, advertise_data.MAX_ENCODED_LENGTH))

        if not scan_pass:
            raise exceptions.InvalidOperationException("Encoded Scan Response data length is too long ({} bytes). "
                                                       "Max: {} bytes".format(scan_len, advertise_data.MAX_ENCODED_LENGTH))

        self.ble_device.ble_driver.ble_gap_adv_data_set(advertise_data.to_ble_adv_data(), scan_response.to_ble_adv_data())

    def set_default_advertise_params(self, advertise_interval_ms, timeout_seconds, advertise_mode=AdvertisingMode.connectable_undirected):
        """
        Sets the default advertising parameters so they do not need to be specified on each start

        :param advertise_interval_ms: The advertising interval, in milliseconds
        :param timeout_seconds: How long to advertise for before timing out, in seconds
        :param advertise_mode: The mode the advertiser should use
        :type advertise_mode: AdvertisingMode
        """
        self._advertise_interval = advertise_interval_ms
        self._timeout = timeout_seconds
        self._advertise_mode = advertise_mode

    def start(self, adv_interval_ms=None, timeout_sec=None, auto_restart=None, advertise_mode=None):
        """
        Starts advertising with the given parameters. If none given, will use the default

        :param adv_interval_ms: The interval at which to send out advertise packets, in milliseconds
        :param timeout_sec: The duration which to advertise for
        :param auto_restart: Flag indicating that advertising should restart automatically when the timeout expires, or
                             when the client disconnects
        :param advertise_mode: The mode the advertiser should use
        :return: A waitable that will expire either when the timeout occurs, or a client connects.
                 Waitable Returns a peer.Client() object
        :rtype: ClientConnectionWaitable
        """
        if self._is_advertising:
            self._stop()
        if adv_interval_ms is None:
            adv_interval_ms = self._advertise_interval
        if timeout_sec is None:
            timeout_sec = self._timeout
        if advertise_mode is None:
            advertise_mode = self._advertise_mode
        if auto_restart is None:
            auto_restart = self._auto_restart

        self._timeout = timeout_sec
        self._advertise_interval = adv_interval_ms
        self._advertise_mode = advertise_mode
        self._auto_restart = auto_restart

        params = nrf_types.BLEGapAdvParams(adv_interval_ms, timeout_sec, advertise_mode)

        logger.info("Starting advertising, params: {}, auto-restart: {}".format(params, auto_restart))
        self.ble_device.ble_driver.ble_gap_adv_start(params, self._conn_tag)
        self._is_advertising = True
        return ClientConnectionWaitable(self.ble_device, self.client)

    def stop(self):
        """
        Stops advertising and disables the auto-restart functionality (if enabled)
        """
        self._auto_restart = False
        self._stop()

    def _stop(self):
        if not self._is_advertising:
            return
        self._is_advertising = False
        try:
            self.ble_device.ble_driver.ble_gap_adv_stop()
        except Exception:
            pass

    def _handle_adv_timeout(self, driver, event):
        """
        :type event: nrf_events.GapEvtTimeout
        """
        if event.src == nrf_events.BLEGapTimeoutSrc.advertising:
            self._is_advertising = False
            self._on_advertising_timeout.notify(self)
            if self._auto_restart:
                self.start()

    def _handle_connect(self, peer, event):
        self._is_advertising = False

    def _handle_disconnect(self, peer, event):
        if self._auto_restart:
            self.start()



from blatann.waitables.waitable import Waitable
from blatann.nrf.nrf_events import GapEvtConnected, GapEvtTimeout, BLEGapRoles, BLEGapTimeoutSrc, GapEvtDisconnected
from blatann.exceptions import InvalidStateException


class ConnectionWaitable(Waitable):
    def __init__(self, ble_device, current_peer, role=BLEGapRoles.periph):
        """
        :type ble_driver: blatann.device.BleDevice
        :type current_peer: blatann.peer.Peer
        """
        super(ConnectionWaitable, self).__init__()
        self._peer = current_peer
        self._role = role
        self.ble_driver = ble_device.ble_driver
        ble_device.ble_driver.event_subscribe(self._on_connected_event, GapEvtConnected)
        ble_device.ble_driver.event_subscribe(self._on_timeout_event, GapEvtTimeout)

    def wait(self, timeout=None, exception_on_timeout=True):
        """
        :rtype: blatann.peer.Peer
        """
        return super(ConnectionWaitable, self).wait(timeout, exception_on_timeout)

    def _event_occured(self, ble_driver, result):
        ble_driver.event_unsubscribe(self._on_connected_event, GapEvtConnected)
        ble_driver.event_unsubscribe(self._on_timeout_event, GapEvtTimeout)
        self._notify(result)

    def _on_timeout(self):
        self.ble_driver.event_unsubscribe(self._on_connected_event, GapEvtConnected)
        self.ble_driver.event_unsubscribe(self._on_timeout_event, GapEvtTimeout)

    def _on_timeout_event(self, ble_driver, event):
        """
        :type event: GapEvtTimeout
        """
        if self._role == BLEGapRoles.periph:
            expected_source = BLEGapTimeoutSrc.advertising
        else:
            expected_source = BLEGapTimeoutSrc.conn
        if event.src == expected_source:
            self._event_occured(ble_driver, None)

    def _on_connected_event(self, ble_driver, event):
        """
        :type event: GapEvtConnected
        """
        if event.role != self._role:
            return
        self._event_occured(ble_driver, self._peer)


class ClientConnectionWaitable(ConnectionWaitable):
    def __init__(self, ble_device, peer):
        super(ClientConnectionWaitable, self).__init__(ble_device, peer, BLEGapRoles.periph)

    def wait(self, timeout=None, exception_on_timeout=True):
        """
        :rtype: blatann.peer.Client
        """
        return super(ClientConnectionWaitable, self).wait(timeout, exception_on_timeout)


class PeripheralConnectionWaitable(ConnectionWaitable):
    def __init__(self, ble_device, peer):
        super(PeripheralConnectionWaitable, self).__init__(ble_device, peer, BLEGapRoles.central)

    def wait(self, timeout=None, exception_on_timeout=True):
        """
        :rtype: blatann.peer.Peripheral
        """
        return super(PeripheralConnectionWaitable, self).wait(timeout, exception_on_timeout)


class DisconnectionWaitable(Waitable):
    def __init__(self, connected_peer):
        """
        :type ble_device: blatann.device.BleDevice
        :type connected_peer: blatann.peer.Peer
        """
        super(DisconnectionWaitable, self).__init__()
        if not connected_peer:
            raise InvalidStateException("Peer already disconnected")
        connected_peer.on_disconnect.register(self._on_disconnect)

    def _on_disconnect(self, disconnected_peer, event_args):
        disconnected_peer.on_disconnect.deregister(self._on_disconnect)
        self._notify(disconnected_peer, event_args)




from blatann.gatt import gatts
from blatann.uuid import Uuid128

PERIPHERAL_NAME = "Periph Test"


# Miscellaneous service, used for testing
MATH_SERVICE_UUID = Uuid128("deadbeef-0011-2345-6679-ab12ccd4f550")

# Hex conversion characteristic. A central writes a byte sequence to this characteristic
# and the peripheral will convert it to its hex representation. e.g. "0123" -> "30313233"
HEX_CONVERT_CHAR_UUID = MATH_SERVICE_UUID.new_uuid_from_base(0xbeaa)

# Counting characteristic. The peripheral will periodically send out a notification on this characteristic
# With a monotonically-increasing, 4-byte little-endian number
COUNTING_CHAR_UUID = MATH_SERVICE_UUID.new_uuid_from_base("1234")

# Properties for the hex conversion and counting characteristics
HEX_CONVERT_CHAR_PROPERTIES = gatts.GattsCharacteristicProperties(read=True, notify=True, indicate=True, write=True,
                                                                  max_length=128, variable_length=True)
COUNTING_CHAR_PROPERTIES = gatts.GattsCharacteristicProperties(read=False, notify=True, max_length=4,
                                                               variable_length=False)

# Time service, report's the current time. Also demonstrating another wait to create a UUID
_TIME_SERVICE_BASE_UUID = "beef0000-0123-4567-89ab-cdef01234567"
TIME_SERVICE_UUID = Uuid128.combine_with_base("dead", _TIME_SERVICE_BASE_UUID)

# Time characteristic. When read, reports the peripheral's current time in a human-readable string
TIME_CHAR_UUID = TIME_SERVICE_UUID.new_uuid_from_base("dddd")

# Properties for the time characteristic
TIME_CHAR_PROPERTIES = gatts.GattsCharacteristicProperties(read=True, max_length=30, variable_length=True)


"""
This example demonstrates reading and subscribing to a peripheral's Battery service to get
updates on the peripheral's current battery levels.
The operations here are programmed in a procedural manner.

This can be used alongside any peripheral which implements the Battery Service and
advertises the 16-bit Battery Service UUID.
The peripheral_battery_service example can be used with this.
"""
import time
from blatann import BleDevice
from blatann.utils import setup_logger
from blatann.services import battery
from blatann.nrf import nrf_events


logger = setup_logger(level="INFO")


def on_battery_level_update(battery_service, event_args):
    """
    :param battery_service:
    :type event_args: blatann.event_args.DecodedReadCompleteEventArgs
    """
    battery_percent = event_args.value
    logger.info("Battery: {}%".format(battery_percent))


def main(serial_port):
    # Open the BLE Device and suppress spammy log messages
    ble_device = BleDevice(serial_port)
    ble_device.event_logger.suppress(nrf_events.GapEvtAdvReport)
    ble_device.open()

    # Set scan duration for 4 seconds
    ble_device.scanner.set_default_scan_params(timeout_seconds=4)
    logger.info("Scanning for peripherals advertising UUID {}".format(battery.BATTERY_SERVICE_UUID))

    target_address = None
    # Start scan and wait for it to complete
    scan_report = ble_device.scanner.start_scan().wait()
    # Search each peer's advertising data for the Battery Service UUID to be advertised
    for report in scan_report.advertising_peers_found:
        if battery.BATTERY_SERVICE_UUID in report.advertise_data.service_uuid16s:
            target_address = report.peer_address
            break

    if not target_address:
        logger.info("Did not find peripheral advertising battery service")
        return

    # Initiate connection and wait for it to finish
    logger.info("Found match: connecting to address {}".format(target_address))
    peer = ble_device.connect(target_address).wait()
    if not peer:
        logger.warning("Timed out connecting to device")
        return

    logger.info("Connected, conn_handle: {}".format(peer.conn_handle))
    # Initiate service discovery and wait for it to complete
    _, event_args = peer.discover_services().wait(10, exception_on_timeout=False)
    logger.info("Service discovery complete! status: {}".format(event_args.status))

    # Find the battery service within the peer's database
    battery_service = battery.find_battery_service(peer.database)
    if not battery_service:
        logger.info("Failed to find Battery Service in peripheral database")
        peer.disconnect().wait()
        return

    # Read out the battery level
    logger.info("Reading battery level...")
    _, event_args = battery_service.read().wait()
    battery_percent = event_args.value
    logger.info("Battery: {}%".format(battery_percent))

    if battery_service.can_enable_notifications:
        battery_service.on_battery_level_updated.register(on_battery_level_update)
        battery_service.enable_notifications().wait()

        wait_duration = 30
        logger.info("Waiting {} seconds for any battery notifications".format(wait_duration))
        time.sleep(wait_duration)

    # Clean up
    logger.info("Disconnecting from peripheral")
    peer.disconnect().wait()
    ble_device.close()


if __name__ == '__main__':
    main("COM9")



import logging
from threading import Lock

from blatann import peer, exceptions
from blatann.gap import advertising, scanning, default_bond_db
from blatann.gatt import gatts, MTU_SIZE_FOR_MAX_DLE, MTU_SIZE_MINIMUM
from blatann.nrf import nrf_events, nrf_types
from blatann.nrf.nrf_driver import NrfDriver, NrfDriverObserver
from blatann.uuid import Uuid, Uuid16, Uuid128
from blatann.waitables.connection_waitable import PeripheralConnectionWaitable

logger = logging.getLogger(__name__)


class _EventLogger(NrfDriverObserver):
    def __init__(self, ble_driver):
        ble_driver.observer_register(self)
        self._suppressed_events = []
        self._lock = Lock()

    def suppress(self, *nrf_event_types):
        with self._lock:
            for e in nrf_event_types:
                if e not in self._suppressed_events:
                    self._suppressed_events.append(e)

    def on_driver_event(self, nrf_driver, event):
        with self._lock:
            if type(event) not in self._suppressed_events:
                logger.debug("Got NRF Driver event: {}".format(event))


class _UuidManager(object):
    def __init__(self, ble_driver):
        """

        :type ble_driver: pc_ble_driver_py.nrf_driver.NrfDriver
        """
        self.ble_driver = ble_driver
        self.registered_vs_uuids = []

    def register_uuid(self, uuid):
        if isinstance(uuid, Uuid16):
            return  # Don't need to register standard 16-bit UUIDs
        elif isinstance(uuid, Uuid128):
            # Check if the base is already registered. If so, do nothing
            for registered_base in self.registered_vs_uuids:
                if uuid.uuid_base == registered_base.base:
                    uuid.nrf_uuid = nrf_types.BLEUUID(uuid.uuid16, registered_base)
                    return
            # Not registered, create a base
            base = nrf_types.BLEUUIDBase(uuid.uuid_base)
            self.ble_driver.ble_vs_uuid_add(base)
            self.registered_vs_uuids.append(base)
            uuid.nrf_uuid = nrf_types.BLEUUID(uuid.uuid16, base)
        elif isinstance(uuid, nrf_types.BLEUUID):
            self.ble_driver.ble_vs_uuid_add(uuid.base)
            self.registered_vs_uuids.append(uuid.base)
        elif isinstance(uuid, nrf_types.BLEUUIDBase):
            self.ble_driver.ble_vs_uuid_add(uuid)
            self.registered_vs_uuids.append(uuid.base)
        else:
            raise ValueError("uuid must be a 16-bit or 128-bit UUID")

    def nrf_uuid_to_uuid(self, nrf_uuid):
        """
        :type nrf_uuid: BLEUUID
        :rtype: Uuid
        """
        if nrf_uuid.base.type == 0:
            raise ValueError("UUID Not registered: {}".format(nrf_uuid))
        if nrf_uuid.base.type == nrf_types.BLEUUIDBase.BLE_UUID_TYPE_BLE:
            return Uuid16(nrf_uuid.get_value())
        base = None
        for uuid_base in self.registered_vs_uuids:
            if nrf_uuid.base.type == uuid_base.type:
                base = uuid_base

        if base is None:
            raise ValueError("Unable to find registered 128-bit uuid: {}".format(nrf_uuid))
        return Uuid128.combine_with_base(nrf_uuid.value, base.base)


class BleDevice(NrfDriverObserver):
    def __init__(self, comport="COM1", baud=1000000, log_driver_comms=False):
        self.ble_driver = NrfDriver(comport, baud, log_driver_comms)
        self.event_logger = _EventLogger(self.ble_driver)
        self.ble_driver.observer_register(self)
        self.ble_driver.event_subscribe(self._on_user_mem_request, nrf_events.EvtUserMemoryRequest)
        self._ble_configuration = self.ble_driver.ble_enable_params_setup()
        self._default_conn_config = nrf_types.BleConnConfig()

        self.bond_db_loader = default_bond_db.DefaultBondDatabaseLoader()
        self.bond_db = default_bond_db.DefaultBondDatabase()

        self.client = peer.Client(self)
        self.connected_peripherals = {}
        self.connecting_peripheral = None

        self.uuid_manager = _UuidManager(self.ble_driver)
        self.advertiser = advertising.Advertiser(self, self.client, self._default_conn_config.conn_tag)
        self.scanner = scanning.Scanner(self)
        self._db = gatts.GattsDatabase(self, self.client)
        self._default_conn_params = peer.DEFAULT_CONNECTION_PARAMS
        self._att_mtu_max = MTU_SIZE_MINIMUM

    def configure(self, vendor_specific_uuid_count=10, service_changed=False, max_connected_peripherals=1,
                  max_connected_clients=1, max_secured_peripherals=1,
                  attribute_table_size=nrf_types.driver.BLE_GATTS_ATTR_TAB_SIZE_DEFAULT,
                  att_mtu_max_size=MTU_SIZE_FOR_MAX_DLE, device_name=""):
        if self.ble_driver.is_open:
            raise exceptions.InvalidStateException("Cannot configure the BLE device after it has been opened")
        if device_name and isinstance(device_name, str):
            device_name = device_name.encode("utf8")

        self._ble_configuration = nrf_types.BleEnableConfig(vendor_specific_uuid_count, max_connected_clients,
                                                            max_connected_peripherals, max_secured_peripherals,
                                                            service_changed, attribute_table_size, device_name)
        self._default_conn_config.max_att_mtu = att_mtu_max_size

    def open(self, clear_bonding_data=False):
        if clear_bonding_data:
            self.clear_bonding_data()
        else:
            self.bond_db = self.bond_db_loader.load()
        self.ble_driver.open()
        self._default_conn_config.conn_count = self._ble_configuration.central_role_count + self._ble_configuration.periph_role_count
        self.ble_driver.ble_conn_configure(self._default_conn_config)
        self.ble_driver.ble_enable(self._ble_configuration)

    def close(self):
        self.bond_db_loader.save(self.bond_db)
        self.ble_driver.close()

    def __del__(self):
        self.close()

    def clear_bonding_data(self):
        logger.info("Clearing out all bonding information")
        self.bond_db.delete_all()
        self.bond_db_loader.save(self.bond_db)

    @property
    def address(self):
        """
        Gets the MAC address of the BLE device

        :rtype: nrf_types.gap.BLEGapAddr
        """
        return self.ble_driver.ble_gap_addr_get()

    @address.setter
    def address(self, address):
        """
        Sets the new address of the device.

        :note: This cannot be performed while the device is advertising, scanning, or initiating a connection

        :param address: The new address
        :type address: nrf_types.gap.BLEGapAddr
        """
        self.ble_driver.ble_gap_addr_set(address)

    @property
    def database(self):
        """
        Gets the local database instance that is accessed by connected clients

        :return: The local database
        :rtype: gatts.GattsDatabase
        """
        return self._db

    @property
    def max_mtu_size(self):
        """
        The maximum allowed ATT MTU size that was configured for the device

        :rtype: int
        """
        return self._default_conn_config.max_att_mtu

    def connect(self, peer_address, connection_params=None):
        """
        Initiates a connection to a peripheral peer with the specified connection parameters, or uses the default
        connection parameters if not specified. The connection will not be complete until the returned waitable
        either times out or reports the newly connected peer

        :param peer_address: The address of the peer to connect to
        :type peer_address: peer.PeerAddress
        :param connection_params: Optional connection parameters to use. If not specified, uses the set default
        :type connection_params: peer.ConnectionParameters
        :return: A Waitable which can be used to wait until the connection is successful or times out. Waitable returns
                 a peer.Peripheral object
        :rtype: PeripheralConnectionWaitable
        """
        if peer_address in self.connected_peripherals.keys():
            raise exceptions.InvalidStateException("Already connected to {}".format(peer_address))
        if self.connecting_peripheral is not None:
            raise exceptions.InvalidStateException("Cannot initiate a new connection while connecting to another")

        if not connection_params:
            connection_params = self._default_conn_params

        self.connecting_peripheral = peer.Peripheral(self, peer_address, connection_params)
        periph_connection_waitable = PeripheralConnectionWaitable(self, self.connecting_peripheral)
        self.ble_driver.ble_gap_connect(peer_address, conn_params=connection_params,
                                        conn_cfg_tag=self._default_conn_config.conn_tag)
        return periph_connection_waitable

    def set_default_peripheral_connection_params(self, min_interval_ms, max_interval_ms, timeout_ms, slave_latency=0):
        """
        Sets the default connection parameters for all subsequent connection attempts to peripherals.
        Refer to the Bluetooth specifications for the valid ranges

        :param min_interval_ms: The minimum desired connection interval, in milliseconds
        :param max_interval_ms: The maximum desired connection interval, in milliseconds
        :param timeout_ms: The connection timeout period, in milliseconds
        :param slave_latency: The connection slave latency
        """
        self._default_conn_params = peer.ConnectionParameters(min_interval_ms, max_interval_ms,
                                                              timeout_ms, slave_latency)

    def _on_user_mem_request(self, nrf_driver, event):
        # Only action that can be taken
        self.ble_driver.ble_user_mem_reply(event.conn_handle)

    def on_driver_event(self, nrf_driver, event):
        if isinstance(event, nrf_events.GapEvtConnected):
            conn_params = peer.ConnectionParameters(event.conn_params.min_conn_interval_ms,
                                                    event.conn_params.max_conn_interval_ms,
                                                    event.conn_params.conn_sup_timeout_ms,
                                                    event.conn_params.slave_latency)
            if event.role == nrf_events.BLEGapRoles.periph:
                self.client.peer_connected(event.conn_handle, event.peer_addr, conn_params)
            else:
                if self.connecting_peripheral.peer_address != event.peer_addr:
                    logger.warning("Mismatching address between connecting peripheral and peer event: "
                                   "{} vs {}".format(self.connecting_peripheral.address, event.peer_addr))
                else:
                    self.connected_peripherals[self.connecting_peripheral.peer_address] = self.connecting_peripheral
                    self.connecting_peripheral.peer_connected(event.conn_handle, event.peer_addr, conn_params)
                self.connecting_peripheral = None
        if isinstance(event, nrf_events.GapEvtTimeout):
            if event.src == nrf_events.BLEGapTimeoutSrc.conn:
                self.connecting_peripheral = None
        if isinstance(event, nrf_events.GapEvtDisconnected):
            for peer_address, p in self.connected_peripherals.items():
                if p.conn_handle == event.conn_handle:
                    del self.connected_peripherals[peer_address]
                    return



from __future__ import annotations
import datetime
import logging
import pytz.reference

from blatann.event_type import EventSource, Event
from blatann.exceptions import InvalidOperationException
from blatann.services import DecodedReadWriteEventDispatcher, DecodedReadCompleteEventArgs, DecodedWriteEventArgs
from blatann.services.current_time.constants import *
from blatann.services.current_time.data_types import *
from blatann.gatt.gatts import GattsService, GattsCharacteristicProperties
from blatann.waitables.event_waitable import IdBasedEventWaitable, EventWaitable

logger = logging.getLogger(__name__)


class CurrentTimeServer(object):

    def __init__(self, service, is_writable=False,
                 enable_local_time_info_char=False, enable_ref_time_info_char=False):
        """
        :type service: GattsService
        :param is_writable:
        :param enable_local_time_info_char:
        :param enable_ref_time_info_char:
        """
        self._service = service
        self._is_writable = is_writable
        self._has_local_time_info = enable_local_time_info_char
        self._has_ref_time_info = enable_ref_time_info_char
        self._current_time_read_callback = self._on_characteristic_read_auto
        self._time_delta = datetime.timedelta()

        self._on_current_time_write_event = EventSource("Current Time Write Event")
        self._on_local_time_info_write_event = EventSource("Local Time Info Write Event")

        self._current_time_dispatcher = DecodedReadWriteEventDispatcher(self, CurrentTime,
                                                                        self._on_current_time_write_event, logger)
        self._local_time_dispatcher = DecodedReadWriteEventDispatcher(self, LocalTimeInfo,
                                                                      self._on_local_time_info_write_event, logger)

        cur_time_char_props = GattsCharacteristicProperties(read=True, notify=True, write=is_writable,
                                                            variable_length=False, max_length=CurrentTime.encoded_size())
        self._cur_time_char = service.add_characteristic(CURRENT_TIME_CHARACTERISTIC_UUID, cur_time_char_props)
        self._cur_time_char.on_read.register(self._on_current_time_read)
        self._cur_time_char.on_write.register(self._current_time_dispatcher)

        if enable_local_time_info_char:
            local_time_props = GattsCharacteristicProperties(read=True, notify=True, write=is_writable,
                                                             variable_length=False, max_length=LocalTimeInfo.encoded_size())
            self._local_time_char = service.add_characteristic(LOCAL_TIME_INFO_CHARACTERISTIC_UUID, local_time_props)
            self.set_local_time_info()
            self._local_time_char.on_write.register(self._local_time_dispatcher)

        if enable_ref_time_info_char:
            ref_time_props = GattsCharacteristicProperties(read=True, notify=False, write=False,
                                                           variable_length=False, max_length=ReferenceTimeInfo.encoded_size())
            self._ref_time_char = service.add_characteristic(REFERENCE_INFO_CHARACTERISTIC_UUID, ref_time_props)
            self.set_reference_info()

        self.set_time(datetime.datetime.utcfromtimestamp(0))

    def _on_characteristic_read_auto(self):
        return datetime.datetime.now() + self._time_delta

    def _on_current_time_read(self, char, event_args):
        dt = CurrentTime(self._current_time_read_callback())
        self._cur_time_char.set_value(dt.encode())

    def _on_current_time_write(self, char, event_args):
        logger.info(event_args)

    def _on_local_time_write(self, char, event_args):
        logger.info(event_args)

    @property
    def is_writable(self) -> bool:
        """
        Gets whether or not the service was configured to allow writes to the Current Time and Local Time Info
        characteristics
        """
        return self._is_writable

    @property
    def has_local_time_info(self) -> bool:
        """
        Gets whether or not the service was configured to show the Local Time Info characteristic
        """
        return self._has_local_time_info

    @property
    def has_reference_time_info(self) -> bool:
        """
        Gets whether or not the service was configured to show the Reference Time Info characteristic
        """
        return self._has_ref_time_info

    @property
    def on_current_time_write(self) -> Event[CurrentTimeServer, DecodedWriteEventArgs[CurrentTime]]:
        """
        Event that is triggered when a client writes to the Current Time Characteristic.
        Event emits a DecodedWriteEventArgs argument where the value is of type current_time.CurrentTime
        """
        return self._on_current_time_write_event

    @property
    def on_local_time_info_write(self) -> Event[CurrentTimeServer, DecodedWriteEventArgs[LocalTimeInfo]]:
        """
        Event that is triggered when a client writes to the Local Time Info Characteristic (if present).
        Event emits a DecodedWriteEventArgs argument where the value is of type current_time.LocalTimeInfo
        """
        return self._on_local_time_info_write_event

    def configure_automatic(self):
        """
        Configures the current time and local time info (if present) to use the system time
        """
        now = datetime.datetime.now()

        adj_reason = AdjustmentReason(AdjustmentReasonType.manual_time_update,
                                      AdjustmentReasonType.dst_change,
                                      AdjustmentReasonType.time_zone_change)
        self.set_time(adjustment_reason=adj_reason)

        if self.has_local_time_info:
            local_timezone = pytz.reference.LocalTimezone()
            offset = local_timezone.utcoffset(now).total_seconds()
            dst = local_timezone.dst(now).total_seconds()
            # Get the actual offset by subtracting the DST
            offset -= dst
            dst_enum = DaylightSavingsTimeOffset.from_seconds(dst)

            self.set_local_time_info(offset/3600.0, dst_enum)

    def set_time(self, date=None, adjustment_reason=None, characteristic_read_callback=None):
        """
        Manually sets the time to report to the client.

        If characteristic_read_callback is supplied,
        the function is called for future reads on that characteristic to fetch the current time
        If characteristic_read_callback is None,
        future reads will be based off of the base datetime given and the time passed

        :param date: The new base date to set the characteristic to. Future characteristic reads will base its time
                     off of this value if characteristic_read_callback is not supplied.
                     If the date is not supplied, will use the current system time
                     (same as configure_automatic but doesn't configure local time info)
        :type date: datetime.datetime
        :param adjustment_reason: Optional reason to give for the adjustment
        :type adjustment_reason: AdjustmentReason
        :param characteristic_read_callback: Optional callback to fetch subsequent time values.
                                             Function signature should take no parameters and return a datetime object
        """
        if date is None:
            date = datetime.datetime.now()
            self._time_delta = datetime.timedelta()
        else:
            delta = date - datetime.datetime.now()
            if abs(delta.total_seconds()) < 1:
                delta = datetime.timedelta()
            self._time_delta = delta
        if characteristic_read_callback:
            self._current_time_read_callback = characteristic_read_callback
        else:
            self._current_time_read_callback = self._on_characteristic_read_auto

        dt = CurrentTime(date, adjustment_reason)
        self._cur_time_char.set_value(dt.encode(), True)

    def set_local_time_info(self, timezone_hrs=0.0, dst_offset=DaylightSavingsTimeOffset.standard_time):
        """
        Sets the local time info characteristic data. Only valid if has_local_time_info is True

        :param timezone_hrs: The timezone to report, in hours
        :param dst_offset: The daylight savings time offset
        :type dst_offset: DaylightSavingsTimeOffset
        :raises: InvalidOperationException if the service was not configured with the local time info
        """
        if not self.has_local_time_info:
            raise InvalidOperationException("Current Time service was not initialized with local time info")
        lt = LocalTimeInfo(timezone_hrs, dst_offset)
        self._local_time_char.set_value(lt.encode(), True)

    def set_reference_info(self, time_source=TimeSource.unknown, accuracy=TimeAccuracy.unknown,
                           hours_since_update=None):
        """
        Sets the time reference info characteristic data. Only valid if has_reference_time_info is True

        :param time_source: The time source to use
        :type time_source: TimeSource
        :param accuracy: The accuracy to report
        :type accuracy: TimeAccuracy
        :param hours_since_update: The number of hours since time reference has been updated
        :raises: InvalidOperationException if the service was not configured with the reference info
        """
        if not self.has_reference_time_info:
            raise InvalidOperationException("Current Time service was not initialized with reference info")
        ri = ReferenceTimeInfo(time_source, accuracy, hours_since_update)
        self._ref_time_char.set_value(ri.encode(), False)

    @classmethod
    def add_to_database(cls, gatts_database, is_writable=False,
                        enable_local_time_info_char=False, enable_ref_time_info_char=False):
        service = gatts_database.add_service(CURRENT_TIME_SERVICE_UUID)
        return cls(service, is_writable, enable_local_time_info_char, enable_ref_time_info_char)


class CurrentTimeClient(object):

    def __init__(self, gattc_service):
        """
        :type gattc_service: blatann.gatt.gattc.GattcService
        """
        self._service = gattc_service
        self._current_time_char = gattc_service.find_characteristic(CURRENT_TIME_CHARACTERISTIC_UUID)
        self._local_time_info_char = gattc_service.find_characteristic(LOCAL_TIME_INFO_CHARACTERISTIC_UUID)
        self._ref_info_char = gattc_service.find_characteristic(REFERENCE_INFO_CHARACTERISTIC_UUID)
        self._on_current_time_updated_event = EventSource("Current Time Update Event")
        self._on_local_time_info_updated_event = EventSource("Local Time Info Update Event")
        self._on_reference_info_updated_event = EventSource("Reference Info Update Event")

        self._current_time_dispatcher = DecodedReadWriteEventDispatcher(self, CurrentTime,
                                                                        self._on_current_time_updated_event, logger)
        self._local_time_dispatcher = DecodedReadWriteEventDispatcher(self, LocalTimeInfo,
                                                                      self._on_local_time_info_updated_event, logger)
        self._ref_time_dispatcher = DecodedReadWriteEventDispatcher(self, ReferenceTimeInfo,
                                                                    self._on_reference_info_updated_event, logger)

    @property
    def on_current_time_updated(self) -> Event[CurrentTimeClient, DecodedReadCompleteEventArgs[CurrentTime]]:
        """
        Event triggered when the server has updated its current time
        """
        return self._on_current_time_updated_event

    @property
    def on_local_time_info_updated(self) -> Event[CurrentTimeClient, DecodedReadCompleteEventArgs[LocalTimeInfo]]:
        """
        Event triggered when the server has updated its local time info
        """
        return self._on_local_time_info_updated_event

    @property
    def on_reference_info_updated(self) -> Event[CurrentTimeClient, DecodedReadCompleteEventArgs[ReferenceTimeInfo]]:
        """
        Event triggered when the server has updated its reference time info
        """
        return self._on_reference_info_updated_event

    @property
    def has_local_time_info(self) -> bool:
        return self._local_time_info_char is not None

    @property
    def has_reference_info(self) -> bool:
        return self._ref_info_char is not None

    @property
    def can_enable_notifications(self) -> bool:
        return self._current_time_char.subscribable

    @property
    def can_set_current_time(self) -> bool:
        return self._current_time_char.writable

    @property
    def can_set_local_time_info(self) -> bool:
        if not self.has_local_time_info:
            return False
        return self._local_time_info_char.writable

    def read_time(self) -> EventWaitable[CurrentTimeClient, DecodedReadCompleteEventArgs[CurrentTime]]:
        """
        Reads the time from the server
        """
        event = self._current_time_char.read().then(self._current_time_dispatcher)
        return IdBasedEventWaitable(self._on_current_time_updated_event, event.id)

    def set_time(self, date, adjustment_reason=None):
        """
        Sets the time on the server to the datetime provided

        :type date: datetime.datetime
        :type adjustment_reason: AdjustmentReason
        """
        dt = CurrentTime(date, adjustment_reason)
        return self._current_time_char.write(dt.encode())



from enum import IntEnum

from blatann.nrf.nrf_types import *
from blatann.nrf.nrf_dll_load import driver
import blatann.nrf.nrf_driver_types as util
from blatann.nrf.nrf_events.generic_events import BLEEvent


class GapEvt(BLEEvent):
    pass


class GapEvtAdvReport(GapEvt):
    evt_id = driver.BLE_GAP_EVT_ADV_REPORT

    def __init__(self, conn_handle, peer_addr, rssi, adv_type, adv_data):
        # TODO: What? Adv event has conn_handle? Does not compute
        super(GapEvtAdvReport, self).__init__(conn_handle)
        self.peer_addr = peer_addr
        self.rssi = rssi
        self.adv_type = adv_type
        self.adv_data = adv_data

    def get_device_name(self):
        dev_name_list = []
        if BLEAdvData.Types.complete_local_name in self.adv_data.records:
            dev_name_list = self.adv_data.records[BLEAdvData.Types.complete_local_name]
        elif BLEAdvData.Types.short_local_name in self.adv_data.records:
            dev_name_list = self.adv_data.records[BLEAdvData.Types.short_local_name]
        return "".join(map(chr, dev_name_list))

    @classmethod
    def from_c(cls, event):
        adv_report_evt = event.evt.gap_evt.params.adv_report

        if not adv_report_evt.scan_rsp:
            adv_type = BLEGapAdvType(adv_report_evt.type)
        else:
            adv_type = BLEGapAdvType.scan_response

        return cls(conn_handle=event.evt.gap_evt.conn_handle,
                   peer_addr=BLEGapAddr.from_c(adv_report_evt.peer_addr),
                   rssi=adv_report_evt.rssi,
                   adv_type=adv_type,
                   adv_data=BLEAdvData.from_c(adv_report_evt))

    def __repr__(self):
        return "{}(conn_handle={!r}, peer_addr={!r}, rssi={!r}, adv_type={!r}, adv_data={!r})".format(
            self.__class__.__name__, self.conn_handle,
            self.peer_addr, self.rssi, self.adv_type, self.adv_data)


class GapEvtTimeout(GapEvt):
    evt_id = driver.BLE_GAP_EVT_TIMEOUT

    def __init__(self, conn_handle, src):
        super(GapEvtTimeout, self).__init__(conn_handle)
        self.src = src

    @classmethod
    def from_c(cls, event):
        timeout_evt = event.evt.gap_evt.params.timeout
        return cls(conn_handle=event.evt.gap_evt.conn_handle,
                   src=BLEGapTimeoutSrc(timeout_evt.src))

    def __repr__(self):
        return "{}(conn_handle={!r}, src={!r})".format(self.__class__.__name__, self.conn_handle, self.src)


class GapEvtConnParamUpdateRequest(GapEvt):
    evt_id = driver.BLE_GAP_EVT_CONN_PARAM_UPDATE_REQUEST

    def __init__(self, conn_handle, conn_params):
        super(GapEvtConnParamUpdateRequest, self).__init__(conn_handle)
        self.conn_params = conn_params

    @classmethod
    def from_c(cls, event):
        conn_params = event.evt.gap_evt.params.conn_param_update_request.conn_params
        return cls(conn_handle=event.evt.gap_evt.conn_handle,
                   conn_params=BLEGapConnParams.from_c(conn_params))

    def __repr__(self):
        return "{}(conn_handle={!r}, conn_params={!r})".format(self.__class__.__name__, self.conn_handle,
                                                               self.conn_params)


class GapEvtConnParamUpdate(GapEvt):
    evt_id = driver.BLE_GAP_EVT_CONN_PARAM_UPDATE

    def __init__(self, conn_handle, conn_params):
        super(GapEvtConnParamUpdate, self).__init__(conn_handle)
        self.conn_params = conn_params

    @classmethod
    def from_c(cls, event):
        conn_params = event.evt.gap_evt.params.conn_param_update.conn_params
        return cls(conn_handle=event.evt.gap_evt.conn_handle,
                   conn_params=BLEGapConnParams.from_c(conn_params))

    def __repr__(self):
        return "{}(conn_handle={!r}, conn_params={})".format(self.__class__.__name__, self.conn_handle,
                                                             self.conn_params)


class GapEvtConnected(GapEvt):
    evt_id = driver.BLE_GAP_EVT_CONNECTED

    def __init__(self, conn_handle, peer_addr, role, conn_params):
        super(GapEvtConnected, self).__init__(conn_handle)
        self.peer_addr = peer_addr
        self.role = role
        self.conn_params = conn_params

    @classmethod
    def from_c(cls, event):
        connected_evt = event.evt.gap_evt.params.connected
        return cls(conn_handle=event.evt.gap_evt.conn_handle,
                   peer_addr=BLEGapAddr.from_c(connected_evt.peer_addr),
                   role=BLEGapRoles(connected_evt.role),
                   conn_params=BLEGapConnParams.from_c(connected_evt.conn_params))

    def __repr__(self):
        return "{}(conn_handle={!r}, peer_addr={!r}, role={!r}, conn_params={})".format(self.__class__.__name__,
                                                                                        self.conn_handle,
                                                                                        self.peer_addr, self.role,
                                                                                        self.conn_params)


class GapEvtDisconnected(GapEvt):
    evt_id = driver.BLE_GAP_EVT_DISCONNECTED

    def __init__(self, conn_handle, reason):
        super(GapEvtDisconnected, self).__init__(conn_handle)
        self.reason = reason

    @classmethod
    def from_c(cls, event):
        disconnected_evt = event.evt.gap_evt.params.disconnected
        return cls(conn_handle=event.evt.gap_evt.conn_handle,
                   reason=BLEHci(disconnected_evt.reason))

    def __repr__(self):
        return self._repr_format(reason=self.reason)


class GapEvtDataLengthUpdate(GapEvt):
    evt_id = driver.BLE_GAP_EVT_DATA_LENGTH_UPDATE

    def __init__(self, conn_handle, max_tx_octets, max_rx_octets, max_tx_time_us, max_rx_time_us):
        super(GapEvtDataLengthUpdate, self).__init__(conn_handle)
        self.max_tx_octets = max_tx_octets
        self.max_rx_octets = max_rx_octets
        self.max_tx_time_us = max_tx_time_us
        self.max_rx_time_us = max_rx_time_us

    @classmethod
    def from_c(cls, event):
        conn_handle = event.evt.gap_evt.conn_handle
        params = event.evt.gap_evt.params.data_length_update.effective_params
        return cls(conn_handle, params.max_tx_octets, params.max_rx_octets, params.max_tx_time_us, params.max_rx_time_us)

    def __repr__(self):
        return self._repr_format(max_tx_octets=self.max_tx_octets, max_rx_octets=self.max_rx_octets,
                                 max_tx_time_us=self.max_tx_time_us, max_rx_time_us=self.max_rx_time_us)


class GapEvtDataLengthUpdateRequest(GapEvt):
    evt_id = driver.BLE_GAP_EVT_DATA_LENGTH_UPDATE_REQUEST

    def __init__(self, conn_handle, max_tx_octets, max_rx_octets, max_tx_time_us, max_rx_time_us):
        super(GapEvtDataLengthUpdateRequest, self).__init__(conn_handle)
        self.max_tx_octets = max_tx_octets
        self.max_rx_octets = max_rx_octets
        self.max_tx_time_us = max_tx_time_us
        self.max_rx_time_us = max_rx_time_us

    @classmethod
    def from_c(cls, event):
        conn_handle = event.evt.gap_evt.conn_handle
        params = event.evt.gap_evt.params.data_length_update_request.peer_params
        return cls(conn_handle, params.max_tx_octets, params.max_rx_octets, params.max_tx_time_us, params.max_rx_time_us)

    def __repr__(self):
        return self._repr_format(max_tx_octets=self.max_tx_octets, max_rx_octets=self.max_rx_octets,
                                 max_tx_time_us=self.max_tx_time_us, max_rx_time_us=self.max_rx_time_us)


class GapEvtPhyUpdate(GapEvt):
    evt_id = driver.BLE_GAP_EVT_PHY_UPDATE

    def __init__(self, conn_handle, status, tx_phy, rx_phy):
        super(GapEvtPhyUpdate, self).__init__(conn_handle)
        self.status = status
        self.tx_phy = tx_phy
        self.rx_phy = rx_phy

    @classmethod
    def from_c(cls, event):
        conn_handle = event.evt.gap_evt.conn_handle
        params = event.evt.gap_evt.params.phy_update
        try:
            status = BLEHci(params.status)
        except:
            status = params.status
        return cls(conn_handle, status, BLEGapPhy(params.tx_phy), BLEGapPhy(params.rx_phy))

    def __repr__(self):
        return self._repr_format(status=self.status, tx_phy=self.tx_phy, rx_phy=self.rx_phy)


class GapEvtPhyUpdateRequest(GapEvt):
    evt_id = driver.BLE_GAP_EVT_PHY_UPDATE_REQUEST

    def __init__(self, conn_handle, tx_phy, rx_phy):
        super(GapEvtPhyUpdateRequest, self).__init__(conn_handle)
        self.tx_phy = tx_phy
        self.rx_phy = rx_phy

    @classmethod
    def from_c(cls, event):
        conn_handle = event.evt.gap_evt.conn_handle
        params = event.evt.gap_evt.params.phy_update_request.peer_preferred_phys
        return cls(conn_handle, BLEGapPhy(params.tx_phys), BLEGapPhy(params.rx_phys))

    def __repr__(self):
        return self._repr_format(tx_phy=self.tx_phy, rx_phy=self.rx_phy)


import datetime
from enum import IntEnum
from blatann.services import ble_data_types

# See https://www.bluetooth.com/specifications/gatt/viewer?attributeXmlFile=org.bluetooth.service.current_time.xml
# For more info about the data types and values defined here


class DaylightSavingsTimeOffset(IntEnum):
    standard_time = 0
    half_hour_dst = 2
    full_hour_dst = 4
    two_hour_dst = 8
    unknown = 255

    @staticmethod
    def from_seconds(seconds):
        """
        Converts the DST offset in seconds to one of the above enums.
        Values which do not map directly to an above enum will be mapped to unknown.
        Valid values are essentially 0, 1800 (1/2 hr), 3600 (1 hr), and 7200 (2 hr)

        :param seconds: DST offset in seconds
        :return: The corresponding enum value
        """
        # Figure out what DST enum to use
        dst_15_min_incrs = int((seconds / 3600.0) * 4)
        try:
            return DaylightSavingsTimeOffset(dst_15_min_incrs)
        except:
            # Cannot convert to one of the standard enums, use unknown
            return DaylightSavingsTimeOffset.unknown


class AdjustmentReasonType(IntEnum):
    manual_time_update = 0
    external_time_reference_update = 1
    time_zone_change = 2
    dst_change = 3


class TimeSource(IntEnum):
    unknown = 0
    network_time_protocol = 1
    gps = 2
    radio_time_signal = 3
    manual = 4
    atomic_clock = 5
    cellular_network = 6


class TimeAccuracy(IntEnum):
    out_of_range = 254
    unknown = 255


class AdjustmentReason(ble_data_types.Bitfield):
    bitfield_width = 8
    bitfield_enum = AdjustmentReasonType

    def __init__(self, *adjustment_reasons):
        """
        :param adjustment_reasons: The list of reasons for the time adjustment
        :type adjustment_reasons: AdjustmentReasonType
        """
        self.manual_time_update = AdjustmentReasonType.manual_time_update in adjustment_reasons
        self.external_time_reference_update = AdjustmentReasonType.external_time_reference_update in adjustment_reasons
        self.time_zone_change = AdjustmentReasonType.time_zone_change in adjustment_reasons
        self.dst_change = AdjustmentReasonType.dst_change in adjustment_reasons

        super(AdjustmentReason, self).__init__()


class ExactTime256(ble_data_types.BleCompoundDataType):
    data_stream_types = [ble_data_types.DayDateTime, ble_data_types.Uint8]

    def __init__(self, date):
        """
        :type date: datetime.datetime
        """
        self.datetime = date

    def encode(self):
        stream = ble_data_types.DayDateTime(self.datetime).encode()
        fraction_sec_256 = self.datetime.microsecond * 256.0 / 1E6
        stream.encode(ble_data_types.Uint8, int(fraction_sec_256))
        return stream

    @classmethod
    def decode(cls, stream):
        dt, fraction_sec_256 = super(ExactTime256, cls).decode(stream)  # type: datetime.datetime, int
        dt += datetime.timedelta(microseconds=(1.0E6/256) * fraction_sec_256)
        return dt


class CurrentTime(ble_data_types.BleCompoundDataType):
    """
    Class used to report the current time and reason for adjustment
    """
    data_stream_types = [ExactTime256, AdjustmentReason]

    def __init__(self, date, adjustment_reason=None):
        """
        :type date: datetime.datetime
        """
        self.datetime = date
        self.adjustment_reason = adjustment_reason or AdjustmentReason()

    def encode(self):
        stream = ExactTime256(self.datetime).encode()
        stream.encode(AdjustmentReason, self.adjustment_reason)
        return stream

    @classmethod
    def decode(cls, stream):
        date, adjustment_reason = super(CurrentTime, cls).decode(stream)
        return CurrentTime(date, adjustment_reason)

    def __repr__(self):
        return "{}({}, {})".format(self.__class__.__name__, self.datetime.isoformat(), self.adjustment_reason)


class LocalTimeInfo(ble_data_types.BleCompoundDataType):
    data_stream_types = [ble_data_types.Int8, ble_data_types.Uint8]

    def __init__(self, timezone_offset_hrs=0.0, dst_offset=DaylightSavingsTimeOffset.standard_time):
        self.timezone_offset = timezone_offset_hrs
        self.dst_offset = dst_offset

    def encode(self):
        # Timezone is in 15-min increments, convert from hours to 15-min increments
        tz_offset = int(self.timezone_offset * 4)
        return self.encode_values(tz_offset, self.dst_offset)

    @classmethod
    def decode(cls, stream):
        tz_offset_15min, dst_offset = super(LocalTimeInfo, cls).decode(stream)
        # Convert from 15min increments to hours
        tz_offset_hrs = tz_offset_15min / 4.0
        try:
            dst_offset = DaylightSavingsTimeOffset(dst_offset)
        except:
            pass
        return LocalTimeInfo(tz_offset_hrs, dst_offset)

    def __repr__(self):
        return "{}(Timezone: {}, DST: {})".format(self.__class__.__name__, self.timezone_offset, str(self.dst_offset))


class ReferenceTimeInfo(ble_data_types.BleCompoundDataType):
    data_stream_types = [ble_data_types.Uint8, ble_data_types.Uint8,
                         ble_data_types.Uint8, ble_data_types.Uint8]

    def __init__(self, source=TimeSource.unknown, accuracy_seconds=TimeAccuracy.unknown, hours_since_update=None):
        self.source = source
        self.accuracy = accuracy_seconds
        self.hours_since_update = hours_since_update

    def encode(self):
        if self.accuracy not in [TimeAccuracy.unknown, TimeAccuracy.out_of_range]:
            # Accuracy is reported in 125ms increments (1/8 second)
            accuracy = int(self.accuracy * 8)
        else:
            # Special enum, leave as-is
            accuracy = self.accuracy

        if self.hours_since_update is None:
            days_since_update = 255
            hours_since_update = 255
        else:
            days_since_update = int(self.hours_since_update / 24)
            hours_since_update = int(self.hours_since_update) % 24
            if days_since_update > 255:
                days_since_update = 255
                hours_since_update = 255

        return self.encode_values(self.source, accuracy, days_since_update, hours_since_update)

    @classmethod
    def decode(cls, stream):
        src, accuracy, days_since_update, hrs_since_update = super(ReferenceTimeInfo, cls).decode(stream)

        if hrs_since_update == 255 or days_since_update == 255:
            hrs_since_update = None
        else:
            hrs_since_update += days_since_update*24

        if accuracy not in [TimeAccuracy.unknown, TimeAccuracy.out_of_range]:
            accuracy = accuracy / 8.0
        else:
            accuracy = TimeAccuracy(accuracy)

        try:
            src = TimeSource(src)
        except:
            pass

        return ReferenceTimeInfo(src, accuracy, hrs_since_update)

    def __repr__(self):
        return "{}({}, Accuracy (sec): {}, last update (hrs): {})".format(self.__class__.__name__,
                                                                          str(self.source), str(self.accuracy),
                                                                          self.hours_since_update)



from blatann.services.battery.service import BatteryServer as _BatteryServer, BatteryClient as _BatteryClient, SecurityLevel
from blatann.services.battery.constants import BATTERY_SERVICE_UUID, BATTERY_LEVEL_CHARACTERISTIC_UUID


def add_battery_service(gatts_database, enable_notifications=False, security_level=SecurityLevel.OPEN):
    """
    Adds a battery service to the given GATT Server database

    :param gatts_database: The database to add the service to
    :type gatts_database: blatann.gatt.gatts.GattsDatabase
    :param enable_notifications: Whether or not the Battery Level characteristic allows notifications
    :param security_level: The security level to use for the service
    :return: The Battery service
    :rtype: _BatteryServer
    """
    return _BatteryServer.add_to_database(gatts_database, enable_notifications, security_level)


def find_battery_service(gattc_database):
    """
    Finds a battery service in the given GATT client database

    :param gattc_database: the GATT client database to search
    :type gattc_database: blatann.gatt.gattc.GattcDatabase
    :return: The Battery service if found, None if not found
    :rtype: _BatteryClient
    """
    return _BatteryClient.find_in_database(gattc_database)



import logging
from threading import RLock
from blatann.services.glucose.data_types import GlucoseMeasurement
from blatann.services.glucose.racp import RacpResponseCode


logger = logging.getLogger(__name__)


class IGlucoseDatabase(object):
    """
    Defines the interface required for the Glucose Service to fetch records and record info
    """
    def first_record(self):
        """
        Gets the first (oldest) record in the database

        :return: The first record in the database, or None if no records in the database
        :rtype: GlucoseMeasurement
        """
        raise NotImplementedError()

    def last_record(self):
        """
        Gets the last (newest) record in the database

        :return: The last record in the database, or None if no records in the database
        :rtype: GlucoseMeasurement
        """
        raise NotImplementedError()

    def record_count(self, min_seq_num=None, max_seq_num=None):
        """
        Gets the number of records between the minimum and maximum sequence numbers provided.
        The min/max limits are inclusive.

        :param min_seq_num: The minimum sequence number to get. If None, no minimum is requested
        :param max_seq_num: The maximum sequence number to get. If None, no maximum is requested
        :return: The number of records that fit the parameters specified
        :rtype: int
        """
        raise NotImplementedError()

    def get_records(self, min_seq_num=None, max_seq_num=None):
        """
        Gets a list of the records between the minimum sequence and maximum sequence numbers provided.
        The min/max limits are inclusive.

        :param min_seq_num: The minimum sequence number to get. If None, no minimum is requested
        :param max_seq_num: The maximum sequence number to get. If None, no maximum is requested
        :return: The list of glucose measurement records that fit the parameters
        :rtype: list[GlucoseMeasurement]
        """
        raise NotImplementedError()

    def delete_records(self, min_seq_num=None, max_seq_num=None):
        """
        Deletes the records between the minimum sequence and maximum sequence numbers provided.
        The min/max limits are inclusive.

        :param min_seq_num: The minimum sequence number to get. If None, no minimum is requested
        :param max_seq_num: The maximum sequence number to get. If None, no maximum is requested
        :return: The response code to send back for the operation
        :rtype: RacpResponseCode
        """
        raise NotImplementedError()


class BasicGlucoseDatabase(IGlucoseDatabase):
    """
    Basic glucose database which simply stores the records in a sorted list, and provides a method for adding
    new records to the database.
    """
    def __init__(self, init_records=None):
        self._records = []  # type: list[GlucoseMeasurement]
        if init_records is not None:
            self._records = init_records
        self._lock = RLock()

    def _sort(self):
        with self._lock:
            self._records = sorted(self._records, key=lambda r: r.sequence_number)

    def _get_records_in_range(self, min_seq_num, max_seq_num):
        with self._lock:
            records = self._records[:]

        if min_seq_num is not None:
            records = filter(lambda r: r.sequence_number >= min_seq_num, records)
        if max_seq_num is not None:
            records = filter(lambda r: r.sequence_number <= max_seq_num, records)
        return records

    def delete_records(self, min_seq_num=None, max_seq_num=None):
        """
        See IGlucoseDatabase
        """
        with self._lock:
            records = self._get_records_in_range(min_seq_num, max_seq_num)
            logger.info("Deleting records between {} and {} - seqs: {}".format(min_seq_num, max_seq_num,
                                                                               [r.sequence_number for r in records]))
            for r in records:
                self._records.remove(r)
        return RacpResponseCode.success

    def record_count(self, min_seq_num=None, max_seq_num=None):
        """
        See IGlucoseDatabase
        """
        num_records = len(self._get_records_in_range(min_seq_num, max_seq_num))
        logger.info("Got record count between {} and {} -  {} records".format(min_seq_num, max_seq_num, num_records))
        return num_records

    def get_records(self, min_seq_num=None, max_seq_num=None):
        """
        See IGlucoseDatabase
        """
        records = self._get_records_in_range(min_seq_num, max_seq_num)
        logger.info("Getting records between {} and {} - seqs: {}".format(min_seq_num, max_seq_num,
                                                                          [r.sequence_number for r in records]))
        return records

    def first_record(self):
        """
        See IGlucoseDatabase
        """
        with self._lock:
            if self._records:
                record = self._records[0]
            else:
                record = None

        logger.info("Glucose DB: First record requested: {}".format(record))
        return record

    def last_record(self):
        """
        See IGlucoseDatabase
        """
        with self._lock:
            if self._records:
                record = self._records[-1]
            else:
                record = None

        logger.info("Glucose DB: Last record requested: {}".format(record))
        return record

    def add_record(self, glucose_measurement):
        """
        Adds a record to the database. NOTE: the measurement's sequence number must be unique within the database

        :param glucose_measurement: The measurement to add
        :type glucose_measurement: GlucoseMeasurement
        """
        with self._lock:
            if glucose_measurement.sequence_number in [r.sequence_number for r in self._records]:
                raise ValueError("Database already contains a measurement with sequence number {}".format(glucose_measurement.sequence_number))
            self._records.append(glucose_measurement)
        self._sort()



from enum import Enum
import logging
from blatann.nrf.nrf_dll_load import driver
import blatann.nrf.nrf_driver_types as util
from blatann.nrf.nrf_types.generic import BLEUUID, BLEUUIDBase
from blatann.nrf.nrf_types.smp import *
from blatann.nrf.nrf_types.enums import *

logger = logging.getLogger(__name__)

NoneType = type(None)


BLE_GATT_HANDLE_INVALID = driver.BLE_GATT_HANDLE_INVALID

"""
GATT Classes
"""
# TODO: BleGattCharExtProps


class BleGattEnableParams(object):
    def __init__(self, max_att_mtu=0):
        self.att_mtu = max_att_mtu

    def to_c(self):
        params = driver.ble_gatt_enable_params_t()
        params.att_mtu = self.att_mtu
        return params


class BLEGattCharacteristicProperties(object):
    def __init__(self, broadcast=False, read=False, write_wo_resp=False,
                 write=False, notify=False, indicate=False, auth_signed_wr=False):
        self.broadcast = broadcast
        self.read = read
        self.write_wo_resp = write_wo_resp
        self.write = write
        self.notify = notify
        self.indicate = indicate
        self.auth_signed_wr = auth_signed_wr

    @classmethod
    def from_c(cls, gattc_char_props):
        return cls(gattc_char_props.broadcast == 1,
                   gattc_char_props.read == 1,
                   gattc_char_props.write_wo_resp == 1,
                   gattc_char_props.write == 1,
                   gattc_char_props.notify == 1,
                   gattc_char_props.indicate == 1,
                   gattc_char_props.auth_signed_wr == 1)

    def to_c(self):
        params = driver.ble_gatt_char_props_t()
        params.broadcast = int(self.broadcast)
        params.read = int(self.read)
        params.write_wo_resp = int(self.write_wo_resp)
        params.write = int(self.write)
        params.notify = int(self.notify)
        params.indicate = int(self.indicate)
        params.auth_signed_wr = int(self.auth_signed_wr)
        return params


class BLEGattService(object):
    srvc_uuid = BLEUUID(BLEUUID.Standard.service_primary)

    def __init__(self, uuid, start_handle, end_handle):
        self.uuid = uuid
        self.start_handle = start_handle
        self.end_handle = end_handle
        self.chars = []

    @classmethod
    def from_c(cls, gattc_service):
        return cls(uuid=BLEUUID.from_c(gattc_service.uuid),
                   start_handle=gattc_service.handle_range.start_handle,
                   end_handle=gattc_service.handle_range.end_handle)

    def char_add(self, char):
        char.end_handle = self.end_handle
        self.chars.append(char)
        if len(self.chars) > 1:
            self.chars[-2].end_handle = char.handle_decl - 1

    def __repr__(self):
        return "{}(uuid={}, start_handle={}, end_handle={})".format(self.__class__.__name__, self.uuid,
                                                                    self.start_handle, self.end_handle)


class BLEGattCharacteristic(object):
    char_uuid = BLEUUID(BLEUUID.Standard.characteristic)

    def __init__(self, uuid, handle_decl, handle_value, data_decl=None, data_value=None, char_props=None):
        self.uuid = uuid
        self.handle_decl = handle_decl
        self.handle_value = handle_value
        self.data_decl = data_decl
        self.data_value = data_value
        self.char_props = char_props  # TODO: if None, parse first byte of data_decl?
        self.end_handle = None
        self.descs = list()

    @classmethod
    def from_c(cls, gattc_char):
        return cls(uuid=BLEUUID.from_c(gattc_char.uuid),
                   handle_decl=gattc_char.handle_decl,
                   handle_value=gattc_char.handle_value,
                   char_props=BLEGattCharacteristicProperties.from_c(gattc_char.char_props))

    def __repr__(self):
        return "{}(uuid={}, handle_decl={}, handle_value={}, data_decl={}, data_value={}, end_handle={})".format(
            self.__class__.__name__, self.uuid, self.handle_decl, self.handle_value,
            self.data_decl, self.data_value, self.end_handle
        )


class BleGattHandle(object):
    def __init__(self, handle=BLE_GATT_HANDLE_INVALID):
        self.handle = handle


"""
GATTC Classes
"""


class BLEGattcWriteParams(object):
    def __init__(self, write_op, flags, handle, data, offset):
        assert isinstance(write_op, BLEGattWriteOperation), 'Invalid argument type'
        assert isinstance(flags, BLEGattExecWriteFlag), 'Invalid argument type'
        self.write_op = write_op
        self.flags = flags
        self.handle = handle
        self.data = data
        self.offset = offset

    @classmethod
    def from_c(cls, gattc_write_params):
        return cls(write_op=BLEGattWriteOperation(gattc_write_params.write_op),
                   flags=gattc_write_params.flags,
                   handle=gattc_write_params.handle,
                   data=util.uint8_array_to_list(gattc_write_params.p_value,
                                                 gattc_write_params.len))

    def to_c(self):
        self.__data_array = util.list_to_uint8_array(self.data)
        write_params = driver.ble_gattc_write_params_t()
        write_params.p_value = self.__data_array.cast()
        write_params.flags = self.flags.value
        write_params.handle = self.handle
        write_params.offset = self.offset
        write_params.len = len(self.data)
        write_params.write_op = self.write_op.value

        return write_params


class BLEGattcDescriptor(object):
    def __init__(self, uuid, handle, data=None):
        self.handle = handle
        self.uuid = uuid
        self.data = data

    @classmethod
    def from_c(cls, gattc_desc):
        return cls(uuid=BLEUUID.from_c(gattc_desc.uuid),
                   handle=gattc_desc.handle)


class BLEGattcAttrInfo16(object):
    def __init__(self, handle, uuid):
        self.handle = handle
        self.uuid = uuid

    @classmethod
    def from_c(cls, attr_info16):
        handle = attr_info16.handle
        uuid = BLEUUID.from_c(attr_info16.uuid)
        return cls(handle, uuid)

    def __repr__(self):
        return "{}(handle={!r}, uuid={})".format(self.__class__.__name__, self.handle, self.uuid)


class BLEGattcAttrInfo128(object):
    def __init__(self, attr_handle, uuid):
        self.handle = attr_handle
        self.uuid = uuid

    @classmethod
    def from_c(cls, attr_info128):
        uuid = BLEUUID.from_uuid128(attr_info128.uuid)
        attr_handle = attr_info128.handle
        return cls(attr_handle, uuid)

    def __repr__(self):
        return "{}(handle={!r}, uuid={})".format(self.__class__.__name__, self.handle, self.uuid)


"""
GATTS Classes
"""
# TODO: BleGattsCharPf


class BleGattsEnableParams(object):
    def __init__(self, service_changed, attribute_table_size):
        assert attribute_table_size % 4 == 0  # attribute table size must be a multiple of 4
        self.service_changed = service_changed
        self.attribute_table_size = attribute_table_size

    def to_c(self):
        params = driver.ble_gatts_enable_params_t()
        params.service_changed = self.service_changed
        params.attr_tab_size = self.attribute_table_size
        return params


class BLEGattsCharHandles(object):
    def __init__(self, value_handle=0, user_desc_handle=0, cccd_handle=0, sccd_handle=0):
        self.value_handle = value_handle
        self.user_desc_handle = user_desc_handle
        self.cccd_handle = cccd_handle
        self.sccd_handle = sccd_handle

    def to_c(self):
        handle_params = driver.ble_gatts_char_handles_t()
        handle_params.value_handle = self.value_handle
        handle_params.user_desc_handle = self.user_desc_handle
        handle_params.cccd_handle = self.cccd_handle
        handle_params.sccd_handle = self.sccd_handle
        return handle_params

    @classmethod
    def from_c(cls, handle_params):
        return cls(handle_params.value_handle,
                   handle_params.user_desc_handle,
                   handle_params.cccd_handle,
                   handle_params.sccd_handle)


class BLEGattsAttribute(object):
    def __init__(self, uuid, attr_metadata, max_len, value=""):
        self.uuid = uuid
        self.attribute_metadata = attr_metadata
        self.max_len = max_len
        self.value = value

    def to_c(self):
        self.__data__array = util.list_to_uint8_array(self.value)
        params = driver.ble_gatts_attr_t()
        params.p_uuid = self.uuid.to_c()
        params.p_attr_md = self.attribute_metadata.to_c()
        params.max_len = self.max_len
        # TODO
        if self.value:
            params.init_len = len(self.value)
            params.init_offs = 0
            params.p_value = self.__data__array.cast()
        return params


class BLEGattsAttrMetadata(object):
    def __init__(self, read_permissions=BLEGapSecModeType.OPEN, write_permissions=BLEGapSecModeType.OPEN,
                 variable_length=False, read_auth=False, write_auth=False):
        self.read_perm = read_permissions
        self.write_perm = write_permissions
        self.vlen = variable_length
        self.read_auth = read_auth
        self.write_auth = write_auth

    def to_c(self):
        params = driver.ble_gatts_attr_md_t()
        params.read_perm = self.read_perm.to_c()
        params.write_perm = self.write_perm.to_c()
        params.vlen = self.vlen
        params.vloc = 1  # STACK
        params.rd_auth = int(self.read_auth)
        params.wr_auth = int(self.write_auth)
        return params

    @classmethod
    def from_c(cls, params):
        read_perm = BLEGapSecMode.from_c(params.read_perm)
        write_perm = BLEGapSecMode.from_c(params.write_perm)
        vlen = params.vlen
        read_auth = bool(params.rd_auth)
        write_auth = bool(params.wr_auth)
        return cls(read_perm, write_perm, vlen, read_auth, write_auth)


class BLEGattsCharMetadata(object):
    def __init__(self, char_props, user_description="", user_description_max_size=0,
                 user_desc_metadata=None, cccd_metadata=None, sccd_metadata=None):
        self.char_props = char_props
        self.user_description = user_description
        self.user_description_max_len = user_description_max_size
        self.user_desc_metadata = user_desc_metadata
        self.cccd_metadata = cccd_metadata
        self.sccd_metadata = sccd_metadata

    def to_c(self):
        params = driver.ble_gatts_char_md_t()
        params.char_props = self.char_props.to_c()
        if self.cccd_metadata:
            params.p_cccd_md = self.cccd_metadata.to_c()
        # if self.sccd_metadata:
        # TODO:
        # if self.user_description:
        #     params.p_char_user_desc = util.list_to_char_array(self.user_description)
        #     params.char_user_desc_size = len(self.user_description)
        # else:
        #     params.char_user_desc_size = 0
        # params.char_user_desc_max_size = self.user_description_max_len
        # if self.user_desc_metadata:
        #     params.p_user_desc_md = self.user_desc_metadata.to_c()
        #     params.p_sccd_md = self.sccd_metadata.to_c()
        return params

    @classmethod
    def from_c(cls, params):
        pass


class BLEGattsAuthorizeParams(object):
    def __init__(self, gatt_status, update, offset=0, data=""):
        assert isinstance(gatt_status, BLEGattStatusCode)
        self.gatt_status = gatt_status
        self.update = update
        self.offset = offset
        self.data = data

    def to_c(self):
        params = driver.ble_gatts_authorize_params_t()
        params.gatt_status = self.gatt_status.value
        params.update = int(self.update)
        params.offset = self.offset

        self.__data_array = util.list_to_uint8_array(self.data)
        params.p_data = self.__data_array.cast()
        params.len = len(self.data)

        return params


class BLEGattsRwAuthorizeReplyParams(object):
    def __init__(self, read=None, write=None):
        assert isinstance(read, (BLEGattsAuthorizeParams, NoneType))
        assert isinstance(write, (BLEGattsAuthorizeParams, NoneType))

        if read is None and write is None:
            raise ValueError("read or write must be set")
        if isinstance(read, BLEGattsAuthorizeParams) and isinstance(write, BLEGattsAuthorizeParams):
            raise ValueError("Both read and write cannot be set at the same time")

        self.read = read
        self.write = write

    def to_c(self):
        params = driver.ble_gatts_rw_authorize_reply_params_t()
        if self.read:
            params.type = driver.BLE_GATTS_AUTHORIZE_TYPE_READ
            params.params.read = self.read.to_c()
        else:
            params.type = driver.BLE_GATTS_AUTHORIZE_TYPE_WRITE
            params.params.write = self.write.to_c()
        return params


class BLEGattsValue(object):
    def __init__(self, value, offset=0):
        self.value = value
        self.offset = offset

    def to_c(self):
        self.__data_array = util.list_to_uint8_array(self.value)
        params = driver.ble_gatts_value_t()
        params.offset = self.offset
        params.len = len(self.value)
        params.p_value = self.__data_array.cast()
        return params

    @classmethod
    def from_c(cls, params):
        offset = params.offset
        value = bytearray(util.uint8_array_to_list(params.p_value, params.len))
        return cls(value, offset)


class BLEGattsHvx(object):
    def __init__(self, char_handle, hvx_type, data, offset=0):
        assert isinstance(hvx_type, BLEGattHVXType)

        self.handle = char_handle
        self.type = hvx_type
        self.offset = offset
        self.data = data

    def to_c(self):
        params = driver.ble_gatts_hvx_params_t()

        self._len_ptr = driver.new_uint16()
        if self.data:
            self.__data_array = util.list_to_uint8_array(self.data)
            params.p_data = self.__data_array.cast()
            driver.uint16_assign(self._len_ptr, len(self.data))
        else:
            driver.uint16_assign(self._len_ptr, 0)

        params.handle = self.handle
        params.type = self.type.value
        params.offset = self.offset
        params.p_len = self._len_ptr  # TODO: Not sure if this works
        return params



from enum import IntEnum
from blatann.services import ble_data_types


class PnpVendorSource(IntEnum):
    bluetooth_sig = 1
    usb_vendor = 2


class PnpId(ble_data_types.BleCompoundDataType):
    data_stream_types = [ble_data_types.Uint8, ble_data_types.Uint16, ble_data_types.Uint16, ble_data_types.Uint16]

    def __init__(self, vendor_id_source, vendor_id, product_id, product_revision):
        super(PnpId, self).__init__()
        self.vendor_id_source = vendor_id_source
        self.vendor_id = vendor_id
        self.product_id = product_id
        self.product_revision = product_revision

    def encode(self):
        return self.encode_values(self.vendor_id_source, self.vendor_id, self.product_id, self.product_revision)

    @classmethod
    def decode(cls, stream):
        vendor_id_source, vendor_id, product_id, product_version = super(PnpId, cls).decode(stream)
        return PnpId(vendor_id_source, vendor_id, product_id, product_version)

    def __repr__(self):
        return "{}(Vendor ID Source: {}, Vendor ID: {}, Product ID: {}, Product Version: {})".format(
            self.__class__.__name__, self.vendor_id_source, self.vendor_id, self.product_id, self.product_revision)


class SystemId(ble_data_types.BleCompoundDataType):
    data_stream_types = [ble_data_types.Uint40, ble_data_types.Uint24]

    def __init__(self, manufacturer_id, organizationally_unique_id):
        super(SystemId, self).__init__()
        self.manufacturer_id = manufacturer_id
        self.organizationally_unique_id = organizationally_unique_id

    def encode(self):
        """
        :rtype: ble_data_types.BleDataStream
        """
        return self.encode_values(self.manufacturer_id, self.organizationally_unique_id)

    @classmethod
    def decode(cls, stream):
        manufacturer_id, organizationally_unique_id = super(SystemId, cls).decode(stream)
        return SystemId(manufacturer_id, organizationally_unique_id)

    def __repr__(self):
        return "{}(Manufacturer ID: {}, OUI: {})".format(self.__class__.__name__, self.manufacturer_id, self.organizationally_unique_id)



"""
This example implements Nordic's custom UART service and demonstrates how to configure the MTU size.
It is configured to use an MTU size based on the Data Length Extensions feature of BLE for maximum throughput.
This is compatible with the peripheral_uart_service example.

This is a simple example which just echos back any data that the client sends to it.
"""
from builtins import input
from blatann import BleDevice
from blatann.nrf import nrf_events
from blatann.utils import setup_logger
from blatann.services import nordic_uart
from blatann.gatt import MTU_SIZE_FOR_MAX_DLE


logger = setup_logger(level="DEBUG")


def on_connect(peer, event_args):
    """
    Event callback for when a central device connects to us

    :param peer: The peer that connected to us
    :type peer: blatann.peer.Client
    :param event_args: None
    """
    if peer:
        logger.info("Connected to peer")
    else:
        logger.warning("Connection timed out")


def on_disconnect(peer, event_args):
    """
    Event callback for when the client disconnects from us (or when we disconnect from the client)

    :param peer: The peer that disconnected
    :type peer: blatann.peer.Client
    :param event_args: The event args
    :type event_args: blatann.event_args.DisconnectionEventArgs
    """
    logger.info("Disconnected from peer, reason: {}".format(event_args.reason))


def on_mtu_size_update(peer, event_args):
    """
    Callback for when the peer's MTU size has been updated/negotiated

    :param peer: The peer the MTU was updated on
    :type peer: blatann.peer.Client
    :type event_args: blatann.event_args.MtuSizeUpdatedEventArgs
    """
    logger.info("MTU size updated from {} to {}".format(event_args.previous_mtu_size, event_args.current_mtu_size))


def on_data_rx(service, data):
    """
    Called whenever data is received on the RX line of the Nordic UART Service

    :param service: the service the data was received from
    :type service: nordic_uart.service.NordicUartClient
    :param data: The data that was received
    :type data: bytes
    """
    logger.info("Received data (len {}): '{}'".format(len(data), data))


def main(serial_port):
    # Open the BLE Device and suppress spammy log messages
    ble_device = BleDevice(serial_port)
    ble_device.event_logger.suppress(nrf_events.GapEvtAdvReport)
    # Configure the BLE device to support MTU sizes which allow the max data length extension PDU size
    # Note this isn't 100% necessary as the default configuration sets the max to this value also
    ble_device.configure(att_mtu_max_size=MTU_SIZE_FOR_MAX_DLE)
    ble_device.open()

    # Set scan duration for 4 seconds
    ble_device.scanner.set_default_scan_params(timeout_seconds=4)
    ble_device.set_default_peripheral_connection_params(10, 30, 4000)
    logger.info("Scanning for peripherals advertising UUID {}".format(nordic_uart.NORDIC_UART_SERVICE_UUID))

    target_address = None
    # Start scan and wait for it to complete
    scan_report = ble_device.scanner.start_scan().wait()
    # Search each peer's advertising data for the Nordic UART Service UUID to be advertised
    for report in scan_report.advertising_peers_found:
        if nordic_uart.NORDIC_UART_SERVICE_UUID in report.advertise_data.service_uuid128s and report.device_name == "Nordic UART Server":
            target_address = report.peer_address
            break

    if not target_address:
        logger.info("Did not find peripheral advertising Nordic UART service")
        return

    # Initiate connection and wait for it to finish
    logger.info("Found match: connecting to address {}".format(target_address))
    peer = ble_device.connect(target_address).wait()
    if not peer:
        logger.warning("Timed out connecting to device")
        return

    logger.info("Connected, conn_handle: {}".format(peer.conn_handle))

    logger.info("Exchanging MTU")
    peer.exchange_mtu(peer.max_mtu_size).wait(10)
    logger.info("MTU Exchange complete, discovering services")

    # Initiate service discovery and wait for it to complete
    _, event_args = peer.discover_services().wait(exception_on_timeout=False)
    logger.info("Service discovery complete! status: {}".format(event_args.status))

    uart_service = nordic_uart.find_nordic_uart_service(peer.database)
    if not uart_service:
        logger.info("Failed to find Nordic UART service in peripheral database")
        peer.disconnect().wait()
        ble_device.close()
        return

    # Initialize the service
    uart_service.initialize().wait(5)
    uart_service.on_data_received.register(on_data_rx)

    while True:
        data = input("Enter data to send to peripheral (q to exit): ")
        if data == "q":
            break
        uart_service.write(data).wait(10)

    peer.disconnect().wait()
    ble_device.close()


if __name__ == '__main__':
    main("COM9")






from blatann.services import ble_data_types

# See https://www.bluetooth.com/specifications/gatt/viewer?attributeXmlFile=org.bluetooth.service.battery_service.xml
# For more info about the data types and values defined here


class BatteryLevel(ble_data_types.Uint8):
    pass



"""
This example shows how to implement a Device Info Service on a peripheral.

This example can be used alongside the central_device_info_service example running on another nordic device,
or using the Nordic nRF Connect app to connect and browse the peripheral's service data
"""
from blatann import BleDevice
from blatann.examples import example_utils
from blatann.gap import advertising
from blatann.services import device_info
from blatann.waitables import GenericWaitable

logger = example_utils.setup_logger(level="DEBUG")


def on_connect(peer, event_args):
    """
    Event callback for when a central device connects to us

    :param peer: The peer that connected to us
    :type peer: blatann.peer.Client
    :param event_args: None
    """
    if peer:
        logger.info("Connected to peer")
    else:
        logger.warning("Connection timed out")


def on_disconnect(peer, event_args):
    """
    Event callback for when the client disconnects from us (or when we disconnect from the client)

    :param peer: The peer that disconnected
    :type peer: blatann.peer.Client
    :param event_args: The event args
    :type event_args: blatann.event_args.DisconnectionEventArgs
    """
    logger.info("Disconnected from peer, reason: {}".format(event_args.reason))


def main(serial_port):
    # Create and open te device
    ble_device = BleDevice(serial_port)
    ble_device.open()

    # Create and add the Device Info Service to the database
    dis = device_info.add_device_info_service(ble_device.database)

    # Set some characteristics in the DIS. The service only contains characteristics which
    # have values set. The other ones are not present
    dis.set_software_revision("14.2.1")
    dis.set_hardware_revision("A")
    dis.set_firmware_revision("1.0.4")
    dis.set_serial_number("AB1234")
    pnp_id = device_info.PnpId(device_info.PnpVendorSource.bluetooth_sig, 0x0058, 0x0002, 0x0013)
    dis.set_pnp_id(pnp_id)

    # Initiate the advertising data. Advertise the name and DIS service UUID
    name = "Peripheral DIS"
    adv_data = advertising.AdvertisingData(local_name=name, service_uuid16s=device_info.DIS_SERVICE_UUID)
    ble_device.advertiser.set_advertise_data(adv_data)

    # Start advertising
    logger.info("Advertising")
    ble_device.client.on_connect.register(on_connect)
    ble_device.client.on_disconnect.register(on_disconnect)
    ble_device.advertiser.start(timeout_sec=0, auto_restart=True)

    # Create a waitable that will never fire, and wait for some time
    w = GenericWaitable()
    w.wait(60*30, exception_on_timeout=False)  # Keep device active for 30 mins

    # Cleanup
    logger.info("Done")
    ble_device.close()


if __name__ == '__main__':
    main("COM8")



import os
import logging
import pickle
import blatann
from blatann.gap.bond_db import BondDatabase, BondDbEntry, BondDatabaseLoader


logger = logging.getLogger(__name__)


default_db_file = os.path.join(os.path.dirname(blatann.__file__), ".user", "bonding_db.pkl")


# TODO 04.16.2019: Replace pickling with something more secure
class DefaultBondDatabaseLoader(BondDatabaseLoader):
    def __init__(self, filename=default_db_file):
        self.filename = filename

    def load(self):
        if not os.path.exists(self.filename):
            return DefaultBondDatabase()
        try:
            with open(self.filename, "rb") as f:
                db = pickle.load(f)
                logger.info("Loaded Bond DB '{}'".format(self.filename))
                return db
        except Exception as e:
            logger.info("Failed to load Bond DB '{}' -  {}:{}".format(self.filename, type(e).__name__, e.message))
            return DefaultBondDatabase()

    def save(self, db):
        dirname = os.path.dirname(default_db_file)
        if not os.path.exists(dirname):
            os.makedirs(dirname)
        with open(self.filename, "wb") as f:
            pickle.dump(db, f)


class DefaultBondDatabase(BondDatabase):
    def __init__(self):
        self._records = []  # type: list[BondDbEntry]
        self.current_id = 0

    def __iter__(self):
        for r in self._records:
            yield r

    def create(self):
        db_entry = BondDbEntry(self.current_id)
        self.current_id += 1
        return db_entry

    def add(self, db_entry):
        if not isinstance(db_entry, BondDbEntry):
            raise ValueError(db_entry)
        # Verify there isn't already an entry with the same ID
        for r in self:
            if r.id == db_entry.id:
                raise ValueError("There already exists an entry with id {}".format(r.id))
        self._records.append(db_entry)

    def update(self, db_entry):
        if not isinstance(db_entry, BondDbEntry):
            raise ValueError(db_entry)
        # TODO

    def delete(self, db_entry):
        if not isinstance(db_entry, BondDbEntry):
            raise ValueError(db_entry)

        for i, entry in enumerate(self._records):
            if entry.id == db_entry.id:
                del self._records[i]
                return

    def delete_all(self):
        self._records = []



from typing import TypeVar, Generic
from enum import Enum
from blatann.nrf.nrf_types import BLEGattStatusCode as GattStatusCode


TDecodedValue = TypeVar("TDecodedValue")


class GattOperationCompleteReason(Enum):
    """
    The reason why a GATT operation completed
    """
    # Operation successful
    SUCCESS = 0
    # Queue of queued operations (notifications,reads, writes) was leared
    QUEUE_CLEARED = 1
    # The client disconnected before the operation completed
    CLIENT_DISCONNECTED = 2
    # The server disconnected before the operation completed
    SERVER_DISCONNECTED = 3
    # The client unsubscribed from the characteristic before the notification was sent
    CLIENT_UNSUBSCRIBED = 4
    # Unknown Failure
    FAILED = 5
    # The peer failed to respond to the ATT operation
    TIMED_OUT = 6


class EventArgs(object):
    """
    Base Event Arguments class
    """
    pass

# Gap Event Args


class DisconnectionEventArgs(EventArgs):
    """
    Event arguments sent when a peer disconnects
    """
    def __init__(self, reason):
        """
        :param reason: The disconnection reason
        :type reason: blatann.gap.HciStatus
        """
        self.reason = reason


# SMP Event Args

class PairingCompleteEventArgs(EventArgs):
    """
    Event arguments when pairing completes, whether it failed or was successful
    """
    def __init__(self, status, security_level):
        """
        :param status: The pairing status
        :type status: blatann.gap.SecurityStatus
        :param security_level: The security level after pairing/bonding
        :type security_level: blatann.gap.smp.SecurityLevel
        """
        self.status = status
        self.security_level = security_level


class SecurityLevelChangedEventArgs(EventArgs):
    def __init__(self, security_level):
        """
        :param security_level: The new security level
        :type security_level: blatann.gap.smp.SecurityLevel
        """
        self.security_level = security_level


class PasskeyEntryEventArgs(EventArgs):
    """
    Event arguments when a passkey needs to be entered by the user
    """
    def __init__(self, key_type, resolve):
        """
        :param key_type: The type of key to be entered (passcode, or out-of-band)
        :type key_type: blatann.gap.AuthenticationKeyType
        :param resolve: The function that should be called to resolve the passkey. If the key type is passcode,
        parameter should be a 6-digit string, or None to cancel
        :type resolve: function
        """
        self.key_type = key_type
        self.resolve = resolve


class PasskeyDisplayEventArgs(EventArgs):
    """
    Event arguments when a passkey needs to be displayed to the user.
    If match_request is set, the user must confirm that the passkeys match on both devices then send back the confirmation

    """
    def __init__(self, passkey, match_request, match_confirm_callback):
        """
        :param passkey: The passkey to display to the user
        :type passkey: str
        :param match_request: Flag indicating whether or not the user needs to confirm that the passkeys match on both devices
        """
        self.passkey = passkey
        self.match_request = match_request
        self._match_confirm_callback = match_confirm_callback

    def match_confirm(self, keys_match):
        """
        If key matching was requested, this function responds with whether or not the keys matched correctly
        :param keys_match: True if the keys matched, False if not
        """
        if self.match_request:
            self._match_confirm_callback(keys_match)


# Gatt Server Event Args

class WriteEventArgs(EventArgs):
    """
    Event arguments for when a client has written to a characteristic on the local database
    """
    def __init__(self, value):
        """
        :param value: The bytes written to the characteristic
        """
        self.value = value


class DecodedWriteEventArgs(EventArgs, Generic[TDecodedValue]):
    """
    Event arguments for when a client has written to a characteristic on the local database
    and the value has been decoded into a data type
    """
    def __init__(self, value: TDecodedValue, raw_value: bytes):
        """
        :param value: The decoded value that was written to the characteristic.
                      This parameter's type depends on the characteristic
        :param raw_value: The raw bytes that were written
        """
        self.value = value
        self.raw_value = raw_value


class SubscriptionStateChangeEventArgs(EventArgs):
    """
    Event arguments for when a client's subscription state has changed
    """
    def __init__(self, subscription_state):
        """
        :type subscription_state: blatann.gatt.SubscriptionState
        """
        self.subscription_state = subscription_state


class NotificationCompleteEventArgs(EventArgs):
    """
    Event arguments for when a notification has been sent to the client from the notification queue
    """
    Reason = GattOperationCompleteReason

    def __init__(self, notification_id, data, reason):
        """
        :param notification_id: The ID of the notification that completed. This will match an ID of a sent notification
        :param data: The data that was sent (or not sent, if failed) to the client
        :param reason: The reason the notification completed
        :type reason: GattOperationCompleteReason
        """
        self.id = notification_id
        self.data = data
        self.reason = reason


# Gatt Client Event Args

class ReadCompleteEventArgs(EventArgs):
    """
    Event arguments for when a read has completed of a peripheral's characteristic
    """
    def __init__(self, read_id, value, status, reason):
        """
        :param read_id: The ID of the read that completed. This will match an id of an initiated read
        :param value: The value read by the characteristic (bytes)
        :param status: The read status
        :type status: blatann.gatt.GattStatusCode
        :param reason: The reason the read completed
        :type reason: GattOperationCompleteReason
        """
        self.id = read_id
        self.value = value
        self.status = status
        self.reason = reason


class WriteCompleteEventArgs(EventArgs):
    """
    Event arguments for when a write has completed on a peripheral's characteristic
    """
    def __init__(self, write_id, value, status, reason):
        """
        :param write_id: the ID of the write that completed. This will match an id of an initiated write
        :param value: The value that was written to the characteristic (bytes)
        :param status: The write status
        :type status: blatann.gatt.GattStatusCode
        :param reason: The reason the write completed
        :type reason: GattOperationCompleteReason
        """
        self.id = write_id
        self.value = value
        self.status = status
        self.reason = reason


class SubscriptionWriteCompleteEventArgs(EventArgs):
    """
    Event arguments for when changing the subscription state of a characteristic completes
    """
    def __init__(self, write_id, value, status, reason):
        """
        :param write_id: the ID of the write that completed. This will match an id of an initiated write
        :param value: The value that was written to the characteristic (bytes)
        :param status: The write status
        :type status: blatann.gatt.GattStatusCode
        :param reason: The reason the write completed
        :type reason: GattOperationCompleteReason
        """
        self.id = write_id
        self.value = value
        self.status = status
        self.reason = reason


class NotificationReceivedEventArgs(EventArgs):
    """
    Event Arguments for when a notification or indication is received from the peripheral
    """
    def __init__(self, value, is_indication):
        """
        :param value: The data sent in the notification.
                      NOTE: This may not contain the whole characteristic data if more than 1 MTU is required.
                      In that case, a separate read will need to be issued on the characteristic
        :param is_indication: Flag indicating if the event was from an indication or notification
        """
        self.value = value
        self.is_indication = is_indication


class DatabaseDiscoveryCompleteEventArgs(EventArgs):
    """
    Event Arguments for when database discovery completes
    """
    def __init__(self, status):
        """
        :param status: The discovery status
        :type status: blatann.gatt.GattStatusCode
        """
        self.status = status


class DecodedReadCompleteEventArgs(ReadCompleteEventArgs, Generic[TDecodedValue]):
    """
    Event Arguments for when a read on a peripheral's characteristic completes and the data stream returned
    is decoded. If unable to decode the value, the bytes read are still returned
    """
    def __init__(self, read_id, value, status, reason, decoded_stream: TDecodedValue = None):
        """
        :param read_complete_event_args: The read complete event args that this wraps
        :type read_complete_event_args: ReadCompleteEventArgs
        :param decoded_stream: The stream which is decoded into an object. This will vary depending on the decoder
        """
        super(DecodedReadCompleteEventArgs, self).__init__(read_id, value, status, reason)
        self.raw_value = value
        if decoded_stream:
            self.value = decoded_stream

    @staticmethod
    def from_notification_complete_event_args(noti_complete_event_args, decoded_stream=None):
        return DecodedReadCompleteEventArgs(0, noti_complete_event_args.value, GattStatusCode.success,
                                            GattOperationCompleteReason.SUCCESS, decoded_stream)

    @staticmethod
    def from_read_complete_event_args(read_complete_event_args, decoded_stream=None):
        return DecodedReadCompleteEventArgs(read_complete_event_args.id, read_complete_event_args.value,
                                            read_complete_event_args.status, read_complete_event_args.reason,
                                            decoded_stream)


class MtuSizeUpdatedEventArgs(EventArgs):
    """
    Event arguments for when the effective MTU size on a connection is updated
    """
    def __init__(self, previous_mtu_size, current_mtu_size):
        self.previous_mtu_size = previous_mtu_size
        self.current_mtu_size = current_mtu_size



"""
This example demonstrates implementing a central BLE connection in a procedural manner. Each bluetooth operation
performed is done sequentially in a linear fashion, and the main context blocks until each operation completes
before moving on to the rest of the program

This is designed to work alongside the peripheral example running on a separate nordic chip
"""
import struct
from blatann import BleDevice
from blatann.gap import smp
from blatann.examples import example_utils, constants
from blatann.nrf import nrf_events

logger = example_utils.setup_logger(level="DEBUG")


def on_counting_char_notification(characteristic, event_args):
    """
    Callback for when a notification is received from the peripheral's counting characteristic.
    The peripheral will periodically notify a monotonically increasing, 4-byte integer. This callback unpacks
    the value and logs it out

    :param characteristic: The characteristic the notification was on (counting characteristic)
    :type characteristic: blatann.gatt.gattc.GattcCharacteristic
    :param event_args: The event arguments
    :type event_args: blatann.event_args.NotificationReceivedEventArgs
    """
    # Unpack as a little-endian, 4-byte integer
    current_count = struct.unpack("<I", event_args.value)[0]
    logger.info("Counting char notification. Curent count: {}".format(current_count))


def main(serial_port):
    # Set the target to the peripheral's advertised name
    target_device_name = constants.PERIPHERAL_NAME

    # Create and open the BLE device (and suppress spammy logs)
    ble_device = BleDevice(serial_port)
    ble_device.event_logger.suppress(nrf_events.GapEvtAdvReport)
    ble_device.open()

    # Set the scanner to scan for 4 seconds
    ble_device.scanner.set_default_scan_params(timeout_seconds=4)

    logger.info("Scanning for '{}'".format(target_device_name))
    target_address = example_utils.find_target_device(ble_device, target_device_name)

    if not target_address:
        logger.info("Did not find target peripheral")
        return

    # Initiate the connection and wait for it to finish
    logger.info("Found match: connecting to address {}".format(target_address))
    peer = ble_device.connect(target_address).wait()
    if not peer:
        logger.warning("Timed out connecting to device")
        return

    # Wait up to 10 seconds for service discovery to complete
    logger.info("Connected, conn_handle: {}".format(peer.conn_handle))
    _, event_args = peer.discover_services().wait(10, exception_on_timeout=False)
    logger.info("Service discovery complete! status: {}".format(event_args.status))

    # Log each service found
    for service in peer.database.services:
        logger.info(service)

    peer.set_connection_parameters(100, 120, 6000)  # Discovery complete, go to a longer connection interval

    # Pair with the peripheral
    def on_passkey_entry(peer, passkey_event_args):
        """
        Callback for when the user is requested to enter a passkey to resume the pairing process.
        Requests the user to enter the passkey and resolves the event with the passkey entered

        :param peer: the peer the passkey is for
        :param passkey_event_args:
        :type passkey_event_args: blatann.event_args.PasskeyEntryEventArgs
        """
        passkey = input("Enter peripheral passkey: ")
        passkey_event_args.resolve(passkey)

    # Setup the security parameters
    peer.security.set_security_params(passcode_pairing=True, io_capabilities=smp.IoCapabilities.KEYBOARD_DISPLAY,
                                      bond=False, out_of_band=False)
    # Register the callback for when a passkey needs to be entered by the user
    peer.security.on_passkey_required.register(on_passkey_entry)
    # Wait up to 60 seconds for the pairing process
    peer.security.pair().wait(60)

    # Find the counting characteristic
    counting_char = peer.database.find_characteristic(constants.COUNTING_CHAR_UUID)
    if counting_char:
        logger.info("Subscribing to the counting characteristic")
        counting_char.subscribe(on_counting_char_notification).wait(5)
    else:
        logger.warning("Failed to find counting characteristic")

    # Find the hex conversion characteristic. This characteristic takes in a bytestream and converts it to its
    # hex representation. e.g. '0123' -> '30313233'
    hex_convert_char = peer.database.find_characteristic(constants.HEX_CONVERT_CHAR_UUID)
    if hex_convert_char:
        # Generate some data ABCDEFG... Then, incrementally send increasing lengths of strings.
        # i.e. first send 'A', then 'AB', then 'ABC'...
        data_to_convert = bytes(ord('A') + i for i in range(12))
        for i in range(len(data_to_convert)):
            data_to_send = data_to_convert[:i+1]
            logger.info("Converting to hex data: '{}'".format(data_to_send))

            # Write the data, waiting up to 5 seconds for the write to complete
            if not hex_convert_char.write(data_to_send).wait(5, False):
                logger.error("Failed to write data, i={}".format(i))
                break

            # Write was successful, when we read the characteristic the peripheral should have converted the string
            # Once again, initiate a read and wait up to 5 seconds for the read to complete
            char, event_args = hex_convert_char.read().wait(5, False)
            logger.info("Hex: '{}'".format(event_args.value.decode("ascii")))
    else:
        logger.warning("Failed to find hex convert char")

    # Clean up
    logger.info("Disconnecting from peripheral")
    peer.disconnect().wait()
    ble_device.close()


if __name__ == '__main__':
    main("COM9")



"""
This example implements Nordic's custom UART service and demonstrates how to configure the MTU size.
It is configured to use an MTU size based on the Data Length Extensions feature of BLE for maximum throughput.
This is compatible with the nRF Connect app (Android version tested) and the central_uart_service example.

This is a simple example which just echos back any data that the client sends to it.
"""
from blatann import BleDevice
from blatann.gap import advertising
from blatann.utils import setup_logger
from blatann.services import nordic_uart
from blatann.gatt import MTU_SIZE_FOR_MAX_DLE
from blatann.waitables import GenericWaitable


logger = setup_logger(level="DEBUG")


def on_connect(peer, event_args):
    """
    Event callback for when a central device connects to us

    :param peer: The peer that connected to us
    :type peer: blatann.peer.Client
    :param event_args: None
    """
    if peer:
        logger.info("Connected to peer")
        peer.exchange_mtu()
    else:
        logger.warning("Connection timed out")


def on_disconnect(peer, event_args):
    """
    Event callback for when the client disconnects from us (or when we disconnect from the client)

    :param peer: The peer that disconnected
    :type peer: blatann.peer.Client
    :param event_args: The event args
    :type event_args: blatann.event_args.DisconnectionEventArgs
    """
    logger.info("Disconnected from peer, reason: {}".format(event_args.reason))


def on_mtu_size_update(peer, event_args):
    """
    Callback for when the peer's MTU size has been updated/negotiated

    :param peer: The peer the MTU was updated on
    :type peer: blatann.peer.Client
    :type event_args: blatann.event_args.MtuSizeUpdatedEventArgs
    """
    logger.info("MTU size updated from {} to {}".format(event_args.previous_mtu_size, event_args.current_mtu_size))


def on_data_rx(service, data):
    """
    Called whenever data is received on the RX line of the Nordic UART Service

    :param service: the service the data was received from
    :type service: nordic_uart.service.NordicUartServer
    :param data: The data that was received
    :type data: bytes
    """
    logger.info("Received data (len {}): '{}'".format(len(data), data))
    logger.info("Echoing data back to client")
    # Echo it back to the client
    service.write(data)


def on_tx_complete(service, event_args):
    logger.info("Write Complete")


def main(serial_port):
    ble_device = BleDevice(serial_port)
    # Configure the BLE device to support MTU sizes which allow the max data length extension PDU size
    # Note this isn't 100% necessary as the default configuration sets the max to this value also
    ble_device.configure(att_mtu_max_size=MTU_SIZE_FOR_MAX_DLE)
    ble_device.open()

    # Create and add the Nordic UART service
    nus = nordic_uart.add_nordic_uart_service(ble_device.database)
    nus.on_data_received.register(on_data_rx)
    nus.on_write_complete.register(on_tx_complete)

    # Register listeners for when the client connects and disconnects
    ble_device.client.on_connect.register(on_connect)
    ble_device.client.on_disconnect.register(on_disconnect)
    ble_device.client.on_mtu_size_updated.register(on_mtu_size_update)

    # Configure the client to prefer the max MTU size
    ble_device.client.preferred_mtu_size = ble_device.max_mtu_size
    ble_device.client.set_connection_parameters(10, 30, 4000)

    # Advertise the service UUID
    adv_data = advertising.AdvertisingData(flags=0x06, local_name="Nordic UART Server")
    scan_data = advertising.AdvertisingData(service_uuid128s=nordic_uart.NORDIC_UART_SERVICE_UUID)

    ble_device.advertiser.set_advertise_data(adv_data, scan_data)

    logger.info("Advertising")

    ble_device.advertiser.start(timeout_sec=0, auto_restart=True)

    # Create a waitable that waits 5 minutes then exits
    w = GenericWaitable()
    try:
        w.wait(5 * 60, exception_on_timeout=False)
    except KeyboardInterrupt:
        pass
    finally:
        ble_device.close()


if __name__ == '__main__':
    main("COM8")
